# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import overload

from qililab.config import logger
from qililab.qprogram.blocks.block import Block
from qililab.qprogram.calibration import Calibration
from qililab.qprogram.decorators import requires_domain
from qililab.qprogram.operations import (
    Acquire,
    AcquireWithCalibratedWeights,
    Measure,
    MeasureWithCalibratedWaveform,
    MeasureWithCalibratedWaveformWeights,
    MeasureWithCalibratedWeights,
    Play,
    PlayWithCalibratedWaveform,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetMarkers,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.structured_program import StructuredProgram
from qililab.qprogram.variable import Domain
from qililab.waveforms import IQPair, Waveform
from qililab.yaml import yaml


@yaml.register_class
class QProgram(StructuredProgram):
    """QProgram is a hardware-agnostic pulse-level programming interface for describing quantum programs.

    This class provides an interface for building quantum programs,
    including defining operations, managing variables, and handling blocks.
    It contains methods for creating, manipulating and controlling
    the execution flow of quantum operations within a program.

    Examples:

        The following example illustrates how to define a Rabi sequence using QProgram.

        .. code-block:: python3

            from qililab import QProgram, Domain, IQPair, Square

            qp = QProgram()

            # Pulse used for changing the state of qubit
            control_wf = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4.0, drag_correction=-2.5)

            # Pulse used for exciting the resonator for readout
            readout_wf = IQPair(I=Square(amplitude=1.0, duration=400), Q=Square(amplitude=0.0, duration=400))

            # Weights used during integration
            weights = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=1.0, duration=2000))

            # Declare a variable
            gain = qp.variable(label="gain", domain=Domain.Voltage)

            # Loop the variable's value over the range [0.0, 1.0]
            with qp.for_loop(variable=gain, start=0.0, stop=1.0, step=0.01):
                # Change the gain output of the drive_bus
                qp.set_gain(bus="drive_bus", gain=gain)

                # Play the control pulse
                qp.play(bus="drive_bus", waveform=control_wf)

                # Sync the buses
                qp.sync()

                # Measure
                qp.measure(bus="readout_bus", waveform=readout_wf, weights=weights)

    """

    def __init__(self) -> None:
        super().__init__()
        self.qblox = self._QbloxInterface(self)
        self.quantum_machines = self._QuantumMachinesInterface(self)

    def __str__(self) -> str:
        def traverse(block: Block):
            string_elements = []
            for element in block.elements:
                string_elements.append(f"{type(element).__name__}:\n")
                for attr_name in vars(element):
                    # ignore uuid, variables. elements, waveforms and weights are handled separately
                    if attr_name in ("_uuid", "variable", "elements", "waveform", "weights"):
                        continue

                    attr_value = getattr(element, attr_name)
                    # Handle UUID checking and append to string_elements
                    if "UUID" not in str(attr_value):
                        string_elements.append(f"\t{attr_name}: {attr_value}\n")
                    else:
                        string_elements.append(f"\t{attr_name}: None\n")  # pragma: no cover

                if isinstance(element, Block):
                    # handle blocks
                    for string_element in traverse(element):
                        string_elements.append(f"\t{string_element}")

                # if not a block, it is asusmed that element is type Operation
                if hasattr(element, "waveform"):
                    waveform_string = (
                        [f"\tWaveform {type(element.waveform).__name__}:\n"]
                        + [f"\t\t{array_line}\n" for array_line in str(element.waveform.envelope()).split("\n")]
                        if isinstance(element.waveform, Waveform)
                        else [f"\tWaveform I {type(element.waveform.I).__name__}:\n"]
                        + [f"\t\t{array_line}\n" for array_line in str(element.waveform.I.envelope()).split("\n")]
                        + [f"\tWaveform Q {type(element.waveform.Q).__name__}):\n"]
                        + [f"\t\t{array_line}\n" for array_line in str(element.waveform.Q.envelope()).split("\n")]
                    )
                    string_elements.extend(waveform_string)

                if hasattr(element, "weights"):
                    string_elements.append(f"\tWeights I {type(element.weights.I).__name__}:\n")
                    string_elements.extend(
                        [f"\t\t{array_element}\n" for array_element in str(element.weights.I.envelope()).split("\n")]
                    )
                    string_elements.append(f"\tWeights Q {type(element.weights.Q).__name__}:\n")
                    string_elements.extend(
                        [f"\t\t{array_element}\n" for array_element in str(element.weights.Q.envelope()).split("\n")]
                    )

            return string_elements

        return "".join(traverse(self._body))

    def has_calibrated_waveforms_or_weights(self) -> bool:
        """Checks if QProgram has named waveforms or weights. These need to be mapped before compiling to hardware-native code.

        Returns:
            bool: True, if QProgram has waveforms or weights that need to be mapped from calibration.
        """

        def traverse(block: Block):
            for element in block.elements:
                if isinstance(element, Block):
                    if traverse(element):
                        return True
                elif isinstance(
                    element,
                    (
                        PlayWithCalibratedWaveform,
                        AcquireWithCalibratedWeights,
                        MeasureWithCalibratedWaveform,
                        MeasureWithCalibratedWeights,
                        MeasureWithCalibratedWaveformWeights,
                    ),
                ):
                    return True
            return False

        return traverse(self.body)

    def with_bus_mapping(self, bus_mapping: dict[str, str]) -> "QProgram":
        """Returns a copy of the QProgram with bus mappings applied.

        Args:
            bus_mapping (dict[str, str]): A dictionary mapping old bus names to new bus names.

        Returns:
            QProgram: A new instance of QProgram with updated bus names.
        """

        def traverse(block: Block):
            for index, element in enumerate(block.elements):
                if isinstance(element, Block):
                    traverse(element)
                elif hasattr(element, "bus"):
                    bus = getattr(element, "bus")
                    if isinstance(bus, str) and bus in bus_mapping:
                        setattr(block.elements[index], "bus", bus_mapping[bus])
                elif hasattr(element, "buses"):
                    buses = getattr(element, "buses")
                    if isinstance(buses, list):
                        setattr(
                            block.elements[index],
                            "buses",
                            [bus_mapping[bus] if bus in bus_mapping else bus for bus in buses],
                        )

        # Copy qprogram so the original remain unaffected
        copied_qprogram = deepcopy(self)

        # Recursively traverse qprogram applying the bus mapping
        traverse(copied_qprogram.body)

        # Apply the mapping to buses property
        for bus in list(copied_qprogram.buses):
            if bus in bus_mapping:
                copied_qprogram.buses.remove(bus)
                copied_qprogram.buses.add(bus_mapping[bus])

        return copied_qprogram

    def with_calibration(self, calibration: Calibration):
        """Apply calibration to the operations within the QProgram.

        This method traverses the elements of the QProgram, replacing any
        named operations with the corresponding calibrated waveforms specified
        in the given Calibration instance.

        Args:
            calibration (Calibration): The calibration data to apply to the operations.

        Returns:
            QProgram: A new instance of QProgram with calibrated operations.
        """

        def traverse(block: Block):
            for index, element in enumerate(block.elements):
                if isinstance(element, Block):
                    traverse(element)
                elif isinstance(element, PlayWithCalibratedWaveform) and calibration.has_waveform(
                    bus=element.bus, name=element.waveform
                ):
                    waveform = calibration.get_waveform(bus=element.bus, name=element.waveform)
                    play_operation = Play(bus=element.bus, waveform=waveform, wait_time=element.wait_time)
                    block.elements[index] = play_operation
                elif isinstance(element, AcquireWithCalibratedWeights) and calibration.has_weights(
                    bus=element.bus, name=element.weights
                ):
                    weights = calibration.get_weights(bus=element.bus, name=element.weights)
                    acquire_operation = Acquire(bus=element.bus, weights=weights, save_adc=element.save_adc)
                    block.elements[index] = acquire_operation
                elif isinstance(element, MeasureWithCalibratedWaveform) and calibration.has_waveform(
                    bus=element.bus, name=element.waveform
                ):
                    waveform = calibration.get_waveform(bus=element.bus, name=element.waveform)
                    measure_operation = Measure(
                        bus=element.bus,
                        waveform=waveform,
                        weights=element.weights,
                        demodulation=element.demodulation,
                        save_adc=element.save_adc,
                    )
                    block.elements[index] = measure_operation
                elif isinstance(element, MeasureWithCalibratedWeights) and calibration.has_weights(
                    bus=element.bus, name=element.weights
                ):
                    weights = calibration.get_weights(bus=element.bus, name=element.weights)
                    measure_operation = Measure(
                        bus=element.bus,
                        waveform=element.waveform,
                        weights=weights,
                        demodulation=element.demodulation,
                        save_adc=element.save_adc,
                    )
                    block.elements[index] = measure_operation
                elif (
                    isinstance(element, MeasureWithCalibratedWaveformWeights)
                    and calibration.has_waveform(bus=element.bus, name=element.waveform)
                    and calibration.has_weights(bus=element.bus, name=element.weights)
                ):
                    waveform = calibration.get_waveform(bus=element.bus, name=element.waveform)
                    weights = calibration.get_weights(bus=element.bus, name=element.weights)
                    measure_operation = Measure(
                        bus=element.bus,
                        waveform=waveform,
                        weights=weights,
                        demodulation=element.demodulation,
                        save_adc=element.save_adc,
                    )
                    block.elements[index] = measure_operation

        copied_qprogram = deepcopy(self)
        traverse(copied_qprogram.body)
        return copied_qprogram

    @overload
    def play(self, bus: str, waveform: Waveform | IQPair) -> None:
        """Play a single waveform or an I/Q pair of waveforms on the bus.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (Waveform | IQPair): A single waveform or an I/Q pair of waveforms
        """

    @overload
    def play(self, bus: str, waveform: str) -> None:
        """Play a named waveform on the bus.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (str): An identifier of a named waveform.
        """

    def play(self, bus: str, waveform: Waveform | IQPair | str) -> None:
        """Play a waveform, IQPair, or calibrated operation on the specified bus.

        This method handles both playing a waveform or IQPair, and playing a
        calibrated operation based on the type of the argument provided.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (Waveform | IQPair | str): The waveform, IQPair, or alias of named waveform to play.
        """
        operation = (
            PlayWithCalibratedWaveform(bus=bus, waveform=waveform)
            if isinstance(waveform, str)
            else Play(bus=bus, waveform=waveform)
        )
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("duration", Domain.Time)
    def wait(self, bus: str, duration: int):
        """Adds a delay on the bus with a specified time.

        Args:
            bus (str): Unique identifier of the bus.
            time (int): Duration of the delay.
        """
        operation = Wait(bus=bus, duration=duration)
        self._active_block.append(operation)
        self._buses.add(bus)

    @overload
    def measure(self, bus: str, waveform: IQPair, weights: IQPair, save_adc: bool = False):
        """Play a pulse and acquire results.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (IQPair): Waveform played during measurement.
            weights (IQPair): Weights used during demodulation/integration.
            save_adc (bool, optional): If ADC data should be saved. Defaults to False.
        """

    @overload
    def measure(self, bus: str, waveform: str, weights: IQPair, save_adc: bool = False):
        """Play a named pulse and acquire results.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (str): Waveform played during measurement.
            weights (IQPair): Weights used during demodulation/integration.
            save_adc (bool, optional): If ADC data should be saved. Defaults to False.
        """

    @overload
    def measure(self, bus: str, waveform: IQPair, weights: str, save_adc: bool = False):
        """Play a named pulse and acquire results.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (IQPair): Waveform played during measurement.
            weights (str): Weights used during demodulation/integration.
            save_adc (bool, optional): If ADC data should be saved. Defaults to False.
        """

    @overload
    def measure(self, bus: str, waveform: str, weights: str, save_adc: bool = False):
        """Play a named pulse and acquire results.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (str): Waveform played during measurement.
            weights (str): Weights used during demodulation/integration.
            save_adc (bool, optional): If ADC data should be saved. Defaults to False.
        """

    def measure(self, bus: str, waveform: IQPair | str, weights: IQPair | str, save_adc: bool = False):
        """Play a pulse and acquire results.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (IQPair): Waveform played during measurement.
            weights (IQPair): Weights used during demodulation/integration.
            save_adc (bool, optional): If ADC data should be saved. Defaults to False.
        """
        operation: (
            Measure
            | MeasureWithCalibratedWaveform
            | MeasureWithCalibratedWeights
            | MeasureWithCalibratedWaveformWeights
        )
        if isinstance(waveform, IQPair) and isinstance(weights, IQPair):
            operation = Measure(bus=bus, waveform=waveform, weights=weights, save_adc=save_adc)
        elif isinstance(waveform, str) and isinstance(weights, IQPair):
            operation = MeasureWithCalibratedWaveform(bus=bus, waveform=waveform, weights=weights, save_adc=save_adc)
        elif isinstance(waveform, IQPair) and isinstance(weights, str):
            operation = MeasureWithCalibratedWeights(bus=bus, waveform=waveform, weights=weights, save_adc=save_adc)
        elif isinstance(waveform, str) and isinstance(weights, str):
            operation = MeasureWithCalibratedWaveformWeights(
                bus=bus, waveform=waveform, weights=weights, save_adc=save_adc
            )
        self._active_block.append(operation)
        self._buses.add(bus)

    def sync(self, buses: list[str] | None = None):
        """Synchronize operations between buses, so the operations following will start at the same time.

        If no buses are given, then the synchronization will involve all buses present in the QProgram.

        Args:
            buses (list[str], optional): List of unique identifiers of the buses. Defaults to None.
        """
        operation = Sync(buses=buses)
        self._active_block.append(operation)
        if buses:
            self._buses.update(buses)

    def reset_phase(self, bus: str):
        """Reset the absolute phase of the NCO associated with the bus.

        Args:
            bus (str): Unique identifier of the bus.
        """
        operation = ResetPhase(bus=bus)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("phase", Domain.Phase)
    def set_phase(self, bus: str, phase: float):
        """Set the absolute phase of the NCO associated with the bus.

        Args:
            bus (str): Unique identifier of the bus.
            phase (float): The new absolute phase of the NCO.
        """
        operation = SetPhase(bus=bus, phase=phase)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("frequency", Domain.Frequency)
    def set_frequency(self, bus: str, frequency: float):
        """Set the frequency of the NCO associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            frequency (float): The new frequency of the NCO.
        """
        operation = SetFrequency(bus=bus, frequency=frequency)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("gain", Domain.Voltage)
    def set_gain(self, bus: str, gain: float):
        """Set the gain of the AWG associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            gain (float): The new gain of the AWG.
        """
        operation = SetGain(bus=bus, gain=gain)
        self._active_block.append(operation)
        self._buses.add(bus)

    @requires_domain("offset_path0", Domain.Voltage)
    @requires_domain("offset_path1", Domain.Voltage)
    def set_offset(self, bus: str, offset_path0: float, offset_path1: float | None = None):
        """Set the gain of the AWG associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            offset_path0 (float): The new offset of the AWG for path0.
            offset_path1 (float): The new offset of the AWG for path1.
        """
        operation = SetOffset(bus=bus, offset_path0=offset_path0, offset_path1=offset_path1)
        self._active_block.append(operation)
        self._buses.add(bus)

    @yaml.register_class
    class _QbloxInterface:
        def __init__(self, qprogram: "QProgram"):
            self.qprogram = qprogram
            self.disable_autosync: bool = False

        def set_markers(self, bus: str, mask: str):
            """Set the markers based on a 4-bit binary mask.

            Args:
                bus (str): Unique identifier of the bus.
                mask (str): A 4-bit mask, where 0 means that the associated marker is open (no signal), and 1 means that the marker is closed (signal).
            """
            operation = SetMarkers(bus=bus, mask=mask)
            self.qprogram._active_block.append(operation)
            self.qprogram._buses.add(bus)

        @overload
        def acquire(self, bus: str, weights: IQPair, save_adc: bool = False):
            """Acquire results based on the given weights.

            Args:
                bus (str): Unique identifier of the bus.
                weights (IQPair): Weights used during acquisition.
            """

        @overload
        def acquire(self, bus: str, weights: str, save_adc: bool = False):
            """Acquire results based on the given weights.

            Args:
                bus (str): Unique identifier of the bus.
                weights (str): Weights used during acquisition.
            """

        def acquire(self, bus: str, weights: IQPair | str, save_adc: bool = False):
            """Acquire results based on the given weights.

            Args:
                bus (str): Unique identifier of the bus.
                weights (IQPair | str): Weights used during acquisition.
            """
            operation = (
                Acquire(bus=bus, weights=weights, save_adc=save_adc)
                if isinstance(weights, IQPair)
                else AcquireWithCalibratedWeights(bus=bus, weights=weights, save_adc=save_adc)
            )
            self.qprogram._active_block.append(operation)
            self.qprogram._buses.add(bus)

        @overload
        def play(self, bus: str, waveform: Waveform | IQPair, wait_time: int) -> None:
            """Play a single waveform or an I/Q pair of waveforms on the bus.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (Waveform | IQPair): A single waveform or an I/Q pair of waveforms
            """

        @overload
        def play(self, bus: str, waveform: str, wait_time: int) -> None:
            """Play a named waveform on the bus.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (str): An identifier of a named waveform.
            """

        def play(self, bus: str, waveform: Waveform | IQPair | str, wait_time: int) -> None:
            """Play a waveform, IQPair, or calibrated operation on the specified bus.

            This method handles both playing a waveform or IQPair, and playing a
            calibrated operation based on the type of the argument provided.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (Waveform | IQPair | str): The waveform, IQPair, or alias of named waveform to play.
                wait_time (int): Overwrite the value of Q1ASM play instruction's wait_time parameter.
            """
            operation = (
                PlayWithCalibratedWaveform(bus=bus, waveform=waveform, wait_time=wait_time)
                if isinstance(waveform, str)
                else Play(bus=bus, waveform=waveform, wait_time=wait_time)
            )
            self.qprogram._active_block.append(operation)
            self.qprogram._buses.add(bus)

    @yaml.register_class
    class _QuantumMachinesInterface:
        def __init__(self, qprogram: "QProgram"):
            self.qprogram = qprogram

        @overload
        def measure(
            self,
            bus: str,
            waveform: IQPair,
            weights: IQPair,
            save_adc: bool = False,
            rotation: float = 0.0,
            demodulation: bool = True,
        ):
            """Play a pulse and acquire results.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (IQPair): Waveform played during measurement.
                weights (IQPair): Weights used during demodulation/integration.
                save_adc (bool, optional): If ADC data should be saved. Defaults to False.
                rotation (float, optional): Angle in radians to rotate the IQ plane during demodulation/integration. Defaults to 0.0
                demodulation (bool, optional): If demodulation is enabled. Defaults to True.
            """

        @overload
        def measure(
            self,
            bus: str,
            waveform: str,
            weights: IQPair,
            save_adc: bool = False,
            rotation: float = 0.0,
            demodulation: bool = True,
        ):
            """Play a named pulse and acquire results.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (str): Waveform played during measurement.
                weights (IQPair): Weights used during demodulation/integration.
                save_adc (bool, optional): If ADC data should be saved. Defaults to False.
                rotation (float, optional): Angle in radians to rotate the IQ plane during demodulation/integration. Defaults to 0.0
                demodulation (bool, optional): If demodulation is enabled. Defaults to True.
            """

        @overload
        def measure(
            self,
            bus: str,
            waveform: IQPair,
            weights: str,
            save_adc: bool = False,
            rotation: float = 0.0,
            demodulation: bool = True,
        ):
            """Play a named pulse and acquire results.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (IQPair): Waveform played during measurement.
                weights (str): Weights used during demodulation/integration.
                save_adc (bool, optional): If ADC data should be saved. Defaults to False.
                rotation (float, optional): Angle in radians to rotate the IQ plane during demodulation/integration. Defaults to 0.0
                demodulation (bool, optional): If demodulation is enabled. Defaults to True.
            """

        @overload
        def measure(
            self,
            bus: str,
            waveform: str,
            weights: str,
            save_adc: bool = False,
            rotation: float = 0.0,
            demodulation: bool = True,
        ):
            """Play a named pulse and acquire results.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (str): Waveform played during measurement.
                weights (str): Weights used during demodulation/integration.
                save_adc (bool, optional): If ADC data should be saved. Defaults to False.
                rotation (float, optional): Angle in radians to rotate the IQ plane during demodulation/integration. Defaults to 0.0
                demodulation (bool, optional): If demodulation is enabled. Defaults to True.
            """

        def measure(
            self,
            bus: str,
            waveform: IQPair | str,
            weights: IQPair | str,
            save_adc: bool = False,
            rotation: float = 0.0,
            demodulation: bool = True,
        ):
            """Play a pulse and acquire results.

            Args:
                bus (str): Unique identifier of the bus.
                waveform (IQPair): Waveform played during measurement.
                weights (IQPair): Weights used during demodulation/integration.
                save_adc (bool, optional): If raw ADC data should be saved. Defaults to False.
                rotation (float, optional): Angle in radians to rotate the IQ plane during demodulation/integration. Defaults to 0.0
                demodulation (bool, optional): If demodulation is enabled. Defaults to True.
            """
            operation: (
                Measure
                | MeasureWithCalibratedWaveform
                | MeasureWithCalibratedWeights
                | MeasureWithCalibratedWaveformWeights
            )
            if isinstance(waveform, IQPair) and isinstance(weights, IQPair):
                operation = Measure(
                    bus=bus,
                    waveform=waveform,
                    weights=weights,
                    save_adc=save_adc,
                    rotation=rotation,
                    demodulation=demodulation,
                )
            elif isinstance(waveform, str) and isinstance(weights, IQPair):
                operation = MeasureWithCalibratedWaveform(
                    bus=bus,
                    waveform=waveform,
                    weights=weights,
                    save_adc=save_adc,
                    rotation=rotation,
                    demodulation=demodulation,
                )
            elif isinstance(waveform, IQPair) and isinstance(weights, str):
                operation = MeasureWithCalibratedWeights(
                    bus=bus,
                    waveform=waveform,
                    weights=weights,
                    save_adc=save_adc,
                    rotation=rotation,
                    demodulation=demodulation,
                )
            elif isinstance(waveform, str) and isinstance(weights, str):
                operation = MeasureWithCalibratedWaveformWeights(
                    bus=bus,
                    waveform=waveform,
                    weights=weights,
                    save_adc=save_adc,
                    rotation=rotation,
                    demodulation=demodulation,
                )
            self.qprogram._active_block.append(operation)
            self.qprogram._buses.add(bus)

    def draw(self, averages_displayed=False):
        """Draw the QProgram using QBlox Compiler

        Args:
            averages_displayed (bool): False means that all loops on the sequencer starting with avg will only loop once, and True shows all iterations.
                                        The default is False.
        """

        from qililab.instruments.qblox.qblox_draw import QbloxDraw
        from qililab.qprogram.qblox_compiler import QbloxCompiler

        qblox_draw = QbloxDraw()
        compiler = QbloxCompiler()
        sequencer = compiler.compile(self)
        result_draw = qblox_draw.draw(sequencer=sequencer, averages_displayed=averages_displayed)
        logger.warning("The drawing feature is currently only supported for QBlox.")
        return result_draw



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Single Instrument Controller class"""
from qililab.instrument_controllers.instrument_controller import InstrumentController


class SingleInstrumentController(InstrumentController):
    """Controller device that only controls one single instrument.

    Args:
        number_available_modules (int): Number of modules available in the Instrument Controller.
                                        In this case, there is only one instrument available.
    """

    number_available_modules: int = 1

    def _set_device_to_all_modules(self):
        """Sets the initialized device to the attached module.
        For a SingleInstrumentController, by default,
        the same controller device driver is set to the Instrument.
        """
        self.modules[0].device = self.device



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PulseShape abstract base class."""

from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Type, TypeVar

import numpy as np

from qililab.typings import PulseShapeName
from qililab.typings.factory_element import FactoryElement
from qililab.utils import Factory

T = TypeVar("T", bound="PulseShape")


@dataclass(frozen=True, eq=True)
class PulseShape(FactoryElement):
    """Pulse shapes describe the shape of the :class:`Pulse`'s envelopes. ``PulseShape`` is their abstract base class.

    Every child of this interface needs to contain an `envelope` and `to/from_dict` methods (for serialization).

    The `envelope` method will create the corresponding array of each shape.

    Derived: :class:`Rectangular`, :class:`Gaussian`, :class:`Drag`, :class:`Cosine` and :class:`SNZ`
    """

    name: PulseShapeName = field(init=False)  #: Name of the pulse shape.

    @abstractmethod
    def envelope(self, duration: int, amplitude: float, resolution: float) -> np.ndarray:
        """Computes the amplitudes of the pulse shape envelope.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse.
            resolution (float): Resolution of the pulse envelope.

        Returns:
            ndarray: Amplitude of the envelope for each time step.
        """

    @classmethod
    def from_dict(cls: Type[T], dictionary: dict) -> T:
        """Loads a `PulseShape` object from a dictionary representation.

        Args:
            dictionary (dict): Dictionary representation of the `PulseShape` object, containing all its attributes.
                If called directly from the Abstract parent class: `PulseShape`, it must include the child class name to instantiate.

        Returns:
            PulseShape: PulseShape loaded class, including the name of the pulse shape and its attributes set.
        """
        # For calls from parent PulseShape:
        if cls is PulseShape:
            shape_class = Factory.get(name=dictionary["name"])
            return shape_class.from_dict(dictionary)

        # For calls from childs directly:
        local_dictionary = deepcopy(dictionary)
        name = local_dictionary.pop("name", None)
        # If the dict name is not the same as the class name, raise an error:
        if name not in [cls.name.value, None]:
            raise ValueError(f"Class: {cls.name.value} to instantiate, does not match the given dict name {name}")
        return cls(**local_dictionary)

    def to_dict(self) -> dict:
        """Returns dictionary representation of the `PulseShape`. This includes the class name an all its attributes.

        Returns:
            dict: Dictionary describing the `PulseShape`, including the name and its attributes.
        """
        return {"name": self.name.value} | {
            attribute: getattr(self, attribute) for attribute in self.__dataclass_fields__.keys()
        }



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from qililab.config import logger


def check_device_initialized(func):
    """
    Function decorator to check if the device has been initialized.

    Args:
        method (Callable): The method to be decorated.

    Raises:
        RuntimeError: If the device has not been initialized.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # Check if the device has been initialized.
        if not self.is_device_active():
            logger.error("Instrument: %s | Device has not been initialized.")
            raise RuntimeError(f"Device of instrument {self.alias} has not been initialized.")

        # Call the original function if the device is initialized.
        return func(self, *args, **kwargs)

    return wrapper


def log_set_parameter(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # Extract parameters based on whether they are passed positionally or by name
        parameter = kwargs.get("parameter") if "parameter" in kwargs else args[0]
        value = kwargs.get("value") if "value" in kwargs else args[1]
        channel_id = kwargs.get("channel_id") if "channel_id" in kwargs else (args[2] if len(args) > 2 else None)

        # Perform logging
        if channel_id is None:
            logger.debug("Instrument: %s | Setting parameter %s to value %s", self.alias, parameter.value, value)
        else:
            logger.debug(
                "Instrument: %s | Setting parameter %s to value %s in channel %s",
                self.alias,
                parameter.value,
                value,
                channel_id,
            )

        # Call the original function
        return func(self, *args, **kwargs)

    return wrapper



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" MiniCircuits Instrument Controller """
from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentControllerSettings
from qililab.instrument_controllers.single_instrument_controller import SingleInstrumentController
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instruments.mini_circuits.attenuator import Attenuator
from qililab.typings import MiniCircuitsDriver
from qililab.typings.enums import ConnectionName, InstrumentControllerName, InstrumentTypeName


@InstrumentControllerFactory.register
class MiniCircuitsController(SingleInstrumentController):
    """MiniCircuits class

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        device (MiniCircuitsDriver): Instance of the qcodes MiniCircuits class.
        settings (MiniCircuitsSettings): Settings of the instrument.
    """

    name = InstrumentControllerName.MINI_CIRCUITS
    device: MiniCircuitsDriver
    modules: Sequence[Attenuator]

    @dataclass
    class MiniCircuitsControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific MiniCircuits Controller."""

        def __post_init__(self):
            super().__post_init__()
            self.connection.name = ConnectionName.TCP_IP

    settings: MiniCircuitsControllerSettings

    def _initialize_device(self):
        """Initialize device attribute to the corresponding device class."""
        self.device = MiniCircuitsDriver(
            name=f"{self.name.value}_{self.alias}",
            address=self.address,
        )

    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""
        for module in self.modules:
            if not isinstance(module, Attenuator):
                raise ValueError(
                    f"Instrument {type(module)} not supported."
                    + f"The only supported instrument is {InstrumentTypeName.MINI_CIRCUITS}"
                )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Arbitrary waveform."""

import numpy as np

from qililab.yaml import yaml

from .waveform import Waveform


@yaml.register_class
class Arbitrary(Waveform):
    """Arbitrary waveform. Creates a waveform with the passed envelope.

    Args:
        samples(np.ndarray): Passed envelope to base the waveform on.

    Examples:
        If you want to create a waveform with an envelope given by:

        .. code-block:: python

            import numpy as np

            samples = np.ones(50)

        You would just need to do:

        .. code-block:: python

            import qililab as ql

            arbitrary_envelope = ql.Arbitrary(samples=samples)
    """

    def __init__(self, samples: np.ndarray):
        super().__init__()
        self.samples = samples

    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Returns the envelope corresponding to the arbitrary waveform.

        Args:
            resolution (int, optional): Pulse resolution. Defaults to 1.

        Returns:
            np.ndarray: Height of the envelope for each time step.
        """
        if resolution == 1:
            return self.samples

        # Calculate the downsampled length
        downsampled_length = len(self.samples) // resolution

        # Check for valid resolution
        if downsampled_length < 1:
            raise ValueError("Resolution is too high compared to the waveform duration.")

        # Reshape the waveform into blocks and compute mean for each block
        reshaped_samples = self.samples[: downsampled_length * resolution].reshape(-1, resolution)
        downsampled = reshaped_samples.mean(axis=1)

        # Normalize the downsampled waveform
        if downsampled.max() != downsampled.min():  # Avoid division by zero
            normalized = (downsampled - downsampled.min()) / (downsampled.max() - downsampled.min())
            normalized = normalized * (self.samples.max() - self.samples.min()) + self.samples.min()
        else:
            # If all values are the same, just return a constant array
            normalized = np.full_like(downsampled, self.samples.min())

        return normalized



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CircuitOptimizer class"""

from copy import deepcopy
from typing import TypeVar

import numpy as np
from qibo import Circuit, gates

from qililab.settings.digital.digital_compilation_settings import DigitalCompilationSettings

from .native_gates import Drag, _GateHandler

T_circuit = TypeVar("T_circuit", Circuit, list[gates.Gate])


class CircuitOptimizer:
    """Optimizes a circuit, cancelling redundant gates.

    Args:
        settings (DigitalCompilationSettings): Object containing the Digital Compilations Settings and the info on chip's physical qubits.
            It can be obtained from the ``digital_compilation_settings`` attribute of a ``Platform`` object.
    """

    def __init__(self, settings: DigitalCompilationSettings):
        self.settings: DigitalCompilationSettings = settings
        """Object containing the digital compilations settings and the info on chip's physical qubits."""

    @staticmethod
    def optimize_gates(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Main method to optimize the gates of a Quantum Circuit before unrolling to native gates.

        The total optimization can/might be expanded in the future to include more complex gate optimization.

        Check public docstring in :meth:`.CircuitTranspiler.optimize_gates()` for more information.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the Qibo circuit to cancel gates.

        Returns:
            list[gates.Gate]: list of the gates of the Qibo circuit, optimized.
        """
        # Add more future methods of optimizing gates here, like 2local optimizations, cliffords ...:

        circuit_gates = CircuitOptimizer.cancel_pairs_of_hermitian_gates(circuit_gates)
        circuit_gates = CircuitOptimizer.remove_redundant_start_controlled_gates(circuit_gates)
        # circuit_gates = cls.2local_gate_optimization(circuit_gates) TODO:
        return circuit_gates

    @staticmethod
    def cancel_pairs_of_hermitian_gates(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Optimizes circuit by cancelling adjacent hermitian gates.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the Qibo circuit to cancel pairs of hermitians.

        Returns:
            list[gates.Gate]: list of the gates of the Qibo circuit, optimized.
        """
        # Initial and final circuit gates lists, from which to, one by one, after checks, pass non-cancelled gates:
        gates_info: list[tuple] = _GateHandler.get_circuit_gates_info(circuit_gates)

        # We want to do the sweep circuit cancelling gates least once always:
        old_gates_info = deepcopy(gates_info)
        new_gates_info = CircuitOptimizer._sweep_circuit_cancelling_pairs_of_hermitian_gates(gates_info)

        # And then keep iterating, sweeping over the circuit (cancelling gates) each time, until there is full sweep without any cancellations:
        while new_gates_info != old_gates_info:
            old_gates_info = deepcopy(new_gates_info)
            new_gates_info = CircuitOptimizer._sweep_circuit_cancelling_pairs_of_hermitian_gates(new_gates_info)

        # Create optimized circuit, from the obtained non-cancelled list:
        return _GateHandler.create_qibo_gates_from_gates_info(new_gates_info)

    def add_phases_from_RZs_and_CZs_to_drags(self, circuit_gates: list[gates.Gate], nqubits: int) -> list[gates.Gate]:
        """This method adds the phases from RZs and CZs gates of the circuit to the next Drag gates.

            - The CZs added phases on the Drags, come from a correction from their calibration, stored on the setting of the CZs.
            - The RZs added phases on the Drags, come from commuting all the RZs all the way to the end of the circuit, so they can be deleted as "virtual Z gates".

        Check public docstring in :meth:`.CircuitTranspiler.add_phases_from_RZs_and_CZs_to_drags()` for more information.

        Args:
            circuit_gates (list[gates.Gate]): list of native gates of the circuit, to pass phases to the Drag gates.
            nqubits (int): Number of qubits of the circuit.

        Returns:
            list[gates.Gate]: list of re-ordered gates
        """
        supported_gates = ["rz", "drag", "cz", "wait", "measure"]
        new_gates = []
        shift = dict.fromkeys(range(nqubits), 0)
        for gate in circuit_gates:
            if gate.name not in supported_gates:
                raise NotImplementedError(f"{gate.name} not part of native supported gates {supported_gates}")
            if isinstance(gate, gates.RZ):
                shift[gate.qubits[0]] += gate.parameters[0]
            # add CZ phase correction
            elif isinstance(gate, gates.CZ):
                gate_settings = self.settings.get_gate(name=gate.__class__.__name__, qubits=gate.qubits)
                control_qubit, target_qubit = gate.qubits
                corrections = next(
                    (
                        event.pulse.options
                        for event in gate_settings
                        if (
                            event.pulse.options is not None
                            and f"q{control_qubit}_phase_correction" in event.pulse.options
                        )
                    ),
                    None,
                )
                if corrections is not None:
                    shift[control_qubit] += corrections[f"q{control_qubit}_phase_correction"]
                    shift[target_qubit] += corrections[f"q{target_qubit}_phase_correction"]
                new_gates.append(gate)
            else:
                # if gate is drag pulse, shift parameters by accumulated Zs
                if isinstance(gate, Drag):
                    # create new drag pulse rather than modify parameters of the old one
                    gate = Drag(gate.qubits[0], gate.parameters[0], gate.parameters[1] - shift[gate.qubits[0]])

                # append gate to optimized list
                new_gates.append(gate)

        return new_gates

    @staticmethod
    def optimize_transpiled_gates(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Main method to optimize the gates of a Quantum Circuit after having unrolled to native gates.

        The total optimization can/might be expanded in the future to include more complex optimizations.

        Check public docstring in :meth:`.CircuitTranspiler.optimize_transpiled_gates()` for more information.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the transpiled circuit, to optimize.

        Returns:
            list[gates.Gate]: list of gates of the transpiled circuit, optimized.
        """
        # Add more optimizations of the transpiled circuit here:
        circuit_gates = CircuitOptimizer.normalize_angles_of_drags(circuit_gates)
        circuit_gates = CircuitOptimizer.bunch_drag_gates(circuit_gates)
        # TODO: Add bunching of Drag Gates for diff phis!
        circuit_gates = CircuitOptimizer.delete_gates_with_no_amplitude(circuit_gates)
        return circuit_gates

    @staticmethod
    def bunch_drag_gates(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Bunches consecutive Drag gates together into a single one.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the transpiled circuit, to bunch drag gates.

        Returns:
            list[gates.Gate]: list of gates of the transpiled circuit, with drag gates bunched."""

        for idx1, drag1 in enumerate(circuit_gates):
            # We are in the last gate, so there is no next one to merge with:
            if idx1 == len(circuit_gates) - 1:
                break

            # Find a Drag gate to merge with:
            if not isinstance(drag1, Drag):
                continue

            for idx2, drag2 in enumerate(circuit_gates[idx1 + 1 :]):
                # Find next gate in the same qubit
                if drag2 is None or drag2.qubits != drag1.qubits:
                    continue

                # If the next gate in the same qubit is not a Drag gate, we can't optimize:
                if not isinstance(drag2, Drag):
                    break

                # If the next gate in the same qubit is a Drag gate, we can merge them:
                new_drag: Drag | None = CircuitOptimizer.merge_consecutive_drags(drag1, drag2)

                # If not successful bunching, we can't merge this gate, go next one:
                if new_drag is None:
                    break

                # If successful bunching, substitute old gates with new one:
                circuit_gates[idx1], drag1 = new_drag, new_drag
                circuit_gates[idx2 + idx1 + 1] = None

        # Remove None values from the list:
        return [gate for gate in circuit_gates if gate is not None]

    @staticmethod
    def merge_consecutive_drags(drag1: Drag, drag2: Drag) -> Drag | None:
        """Merges two consecutive Drag gates into a single one.

        Args:
            drag1 (Drag): First Drag gate.
            drag2 (Drag): Second Drag gate.

        Returns:
            Drag | None: Merged Drag gate. None, if not possible to merge.
        """
        if drag1.qubits != drag2.qubits:
            raise ValueError("Cannot merge Drag gates acting on different qubits.")

        # For the same phi, we just need to add the theta parameters:
        if np.isclose(drag1.parameters[1], drag2.parameters[1]):
            new_normalized_theta = _GateHandler.normalize_angle(drag1.parameters[0] + drag2.parameters[0])
            return Drag(drag1.qubits[0], new_normalized_theta, drag1.parameters[1])

        # For opposite phi, we just need to subtract the theta parameters:
        if np.isclose(drag1.parameters[1], _GateHandler.normalize_angle(drag2.parameters[1] + np.pi)):
            new_normalized_theta = _GateHandler.normalize_angle(drag1.parameters[0] - drag2.parameters[0])
            return Drag(drag1.qubits[0], new_normalized_theta, drag1.parameters[1])

        # TODO: ADD BUNCHING DRAG GATES FOR GENERAL DIFFERENT PHI's!
        return None  # This should return the merged Drag gate, for different phi's!

    @staticmethod
    def normalize_angles_of_drags(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Normalizes angles in the gates of the circuit.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the transpiled circuit, to normalize the angles.

        Returns:
            list[gates.Gate]: list of gates of the transpiled circuit, with normalized angles.
        """
        for gate in circuit_gates:
            if isinstance(gate, Drag):
                gate.parameters = (
                    _GateHandler.normalize_angle(gate.parameters[0]),
                    _GateHandler.normalize_angle(gate.parameters[1]),
                )
        return circuit_gates

    @staticmethod
    def delete_gates_with_no_amplitude(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Deletes gates without amplitude.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the transpiled circuit, to delete gates without amplitude.

        Returns:
            list[gates.Gate]: list of gates of the transpiled circuit, with gates without amplitude deleted."""
        for idx, gate in enumerate(circuit_gates):
            if isinstance(gate, Drag) and np.isclose(gate.parameters[0], 0):
                circuit_gates[idx] = None
        return [gate for gate in circuit_gates if gate]

    @staticmethod
    def _sweep_circuit_cancelling_pairs_of_hermitian_gates(circuit_gates: list[tuple]) -> list[tuple]:
        """Cancels adjacent gates in a circuit.

        Args:
            circuit_gates (list[tuple]): List of gates in the circuit. Where each gate is a tuple of ('name', 'init_args', 'initi_kwargs')

        Returns:
            list[tuple]: List of gates in the circuit, after cancelling adjacent gates. Where each gate is a tuple of ('name', 'init_args', 'initi_kwargs')
        """
        # List of gates, that are available for cancellation:
        hermitian_gates: list = ["H", "X", "Y", "Z", "CNOT", "CZ", "SWAP"]

        output_circuit_gates: list[tuple] = []

        while circuit_gates:  # If original circuit list, is empty or has one gate remaining, we are done:
            if len(circuit_gates) == 1:
                output_circuit_gates.append(circuit_gates[0])
                break

            # Gate of the original circuit, to find a match for:
            gate, gate_args, gate_kwargs = circuit_gates.pop(0)
            gate_qubits = _GateHandler.extract_qubits_from_gate_args(
                gate_args
            )  # Assuming qubits are the first two args

            # If gate is not hermitian (can't be cancelled), add it to the output circuit and continue:
            if gate not in hermitian_gates:
                output_circuit_gates.append((gate, gate_args, gate_kwargs))
                continue

            subend = False
            for i in range(len(circuit_gates)):
                # Next gates, to compare the original with:
                comp_gate, comp_args, comp_kwargs = circuit_gates[i]
                comp_qubits = _GateHandler.extract_qubits_from_gate_args(
                    comp_args
                )  # Assuming qubits are the first two args

                # Simplify duplication, if same gate and qubits found, without any other in between:
                if gate == comp_gate and gate_args == comp_args and gate_kwargs == comp_kwargs:
                    circuit_gates.pop(i)
                    break

                # Add gate, if there is no other gate that acts on the same qubits:
                if i == len(circuit_gates) - 1:
                    output_circuit_gates.append((gate, gate_args, gate_kwargs))
                    break

                # Add gate and leave comparison_gate loop, if we find a gate in common qubit, that prevents contraction:
                for gate_qubit in gate_qubits:
                    if gate_qubit in comp_qubits:
                        output_circuit_gates.append((gate, gate_args, gate_kwargs))
                        subend = True
                        break
                if subend:
                    break

        return output_circuit_gates

    @staticmethod
    def remove_redundant_start_controlled_gates(
        transpiled_circ: T_circuit, gates_to_remove: tuple[gates.Gate] | gates.Gate | None = None
    ) -> T_circuit:
        """Removes redundant controlled gates at the start of the transpiled circuit.

        Args:
            transpiled_circ (Circuit): Transpiled circuit.
            gates_to_remove (tuple[gates.Gate] | gates.Gate, optional): Tuple of gates to remove. Defaults to all controlled gates.

        Returns:
            Circuit | list[gates.Gate] : Transpiled circuit without redundant gates.
        """

        finished_qubits: set = set()
        is_circuit = isinstance(transpiled_circ, Circuit)

        queue: list[gates.Gate] = deepcopy(transpiled_circ.queue) if is_circuit else deepcopy(transpiled_circ)  # type: ignore [attr-defined]
        wire_names = deepcopy(transpiled_circ.wire_names) if is_circuit else None  # type: ignore [attr-defined]

        for idx, gate in enumerate(queue):
            # When all qubits have finished, we stop:
            if is_circuit and len(finished_qubits) == transpiled_circ.nqubits:  # type: ignore [attr-defined]
                break

            # The condition will be, the gate is one of the given, or the gate has a control qubit.
            is_gate_to_remove_OR_has_control_q = (
                isinstance(gate, gates_to_remove) if gates_to_remove else len(gate.control_qubits) != 0
            )
            # If  gate fulfills the condition, and has no gate before, we remove gate:
            if is_gate_to_remove_OR_has_control_q and all(qubit not in finished_qubits for qubit in gate.qubits):
                CircuitOptimizer._make_gate_None_in_queue(gate, idx, queue, wire_names)  # type: ignore[arg-type]
            # If gate not fulfills the condition, or has another non-removed gate before, we stop for that qubit:
            else:
                finished_qubits.update(gate.qubits)

        queue = [x for x in queue if x is not None]

        # If a circuit is originally passed, return a circuit with the wire_names updated:
        if is_circuit:
            return _GateHandler.create_circuit_from_gates(queue, transpiled_circ.nqubits, wire_names)  # type: ignore [attr-defined]
        # If a queue was passed, return a list.
        return queue

    @staticmethod
    def _make_gate_None_in_queue(gate: gates.Gate, idx: int, queue: list[gates.Gate], wire_names: list[int]) -> None:
        """Removes a gate from the circuit queue and updates wire names if the gate is a SWAP gate.

        Args:
            gate (gates.Gate): The gate to be removed from the circuit.
            idx (int): The index of the gate in the queue.
            queue (list[gates.Gate]): The list representing the circuit queue.
            wire_names (list[int]): Wire names of the transpiled circuit.

        Returns:
            None
        """
        queue[idx] = None
        # If we remove a SWAP gate at the start, change wire names:
        if isinstance(gate, gates.SWAP) and wire_names:
            qubits = gate.qubits
            wire_names[qubits[0]], wire_names[qubits[1]] = wire_names[qubits[1]], wire_names[qubits[0]]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QbloxResult class."""

from copy import deepcopy

import numpy as np
import pandas as pd

from qililab.constants import QBLOXRESULT, RUNCARD
from qililab.exceptions import DataUnavailable
from qililab.result.counts import Counts
from qililab.result.result import Result
from qililab.typings.enums import ResultName
from qililab.utils.factory import Factory

from .constants import SCOPE_ACQ_MAX_DURATION
from .qblox_acquisitions_builder import QbloxAcquisitionsBuilder
from .qblox_scope_acquisitions import QbloxScopeAcquisitions


@Factory.register
class QbloxResult(Result):
    """QbloxResult class. Contains the binning acquisition results obtained from the `module.get_acquisitions` method.

    The input to the constructor should be a dictionary with the following structure:

    - integration: integration data.
        - path_0: input path 0 integration result bin list.
        - path_1: input path 1 integration result bin list.
    - threshold: threshold result bin list.
    - valid: list of valid indications per bin.
    - avg_cnt: list of number of averages per bin.

    Args:
        qblox_raw_results (list[dict]): Raw results obtained from a Qblox digitiser.
        integration_lengths (list[int]): Integration lengths used to get the given results.
    """

    name = ResultName.QBLOX

    def __init__(self, qblox_raw_results: list[dict], integration_lengths: list[int]):
        self.qblox_raw_results = qblox_raw_results
        self.integration_lengths = integration_lengths
        self.qblox_scope_acquisitions = QbloxAcquisitionsBuilder.get_scope(
            integration_lengths=integration_lengths, qblox_raw_results=qblox_raw_results
        )
        self.qblox_bins_acquisitions = QbloxAcquisitionsBuilder.get_bins(
            integration_lengths=integration_lengths, qblox_raw_results=qblox_raw_results
        )
        self._qblox_scope_acquisition_copy = deepcopy(self.qblox_scope_acquisitions)
        self.data_dataframe_indices = self.qblox_bins_acquisitions.data_dataframe_indices

    def _demodulated_scope(self, frequency: float, phase_offset: float = 0.0) -> QbloxScopeAcquisitions:
        """Returns the scope acquisitions demodulated in the given frequency with the given phase offset.

        Returns:
            QbloxScopeAcquisitions: demodulated scope acquisitions."""
        if self.qblox_scope_acquisitions is None:
            raise ValueError("Scope acquisitions cannot be demodulated because it doesn't exist.")
        return self.qblox_scope_acquisitions.demodulated(frequency=frequency, phase_offset=phase_offset)

    def _integrated_scope(
        self, scope_acquisitions=None, integrate_from: int = 0, integrate_to: int = SCOPE_ACQ_MAX_DURATION
    ) -> QbloxScopeAcquisitions:
        """Integrated Scope

        Args:
            scope_acquisitions (_type_, optional): _description_. Defaults to None.
            integrate_from (int, optional): _description_. Defaults to 0.
            integrate_to (int, optional): _description_. Defaults to SCOPE_ACQ_MAX_DURATION.

        Raises:
            ValueError: _description_

        Returns:
            QbloxScopeAcquisitions: _description_
        """
        if scope_acquisitions is None:
            scope_acquisitions = self.qblox_scope_acquisitions
        if scope_acquisitions is None:
            raise ValueError("Scope acquisitions cannot be integrated because it doesn't exist.")
        return scope_acquisitions.integrated(integrate_from=integrate_from, integrate_to=integrate_to)

    def acquisitions(self) -> pd.DataFrame:
        """Return acquisition values.

        Returns:
            pd.DataFrame: I, Q, amplitude and phase.
        """
        return self.qblox_bins_acquisitions.acquisitions()

    def acquisitions_scope(
        self,
        demod_freq: float = 0.0,
        demod_phase_offset: float = 0.0,
        integrate: bool = False,
        integration_range: tuple[int, int] = (0, SCOPE_ACQ_MAX_DURATION),
    ) -> tuple[list[float], list[float]]:
        """Acquisitions Scope

        Args:
            demod_freq (float, optional): _description_. Defaults to 0.0.
            demod_phase_offset (float, optional): _description_. Defaults to 0.0.
            integrate (bool, optional): _description_. Defaults to False.
            integration_range (tuple[int, int], optional): _description_. Defaults to (0, SCOPE_ACQ_MAX_DURATION).

        Raises:
            DataUnavailable: Scope data is not available since it was not stored for this acquisition.

        Returns:
            tuple[list[float], list[float]]
        """
        acquisitions = self.qblox_scope_acquisitions
        if acquisitions is None:
            raise DataUnavailable("Scope data is not available since it was not stored for this acquisition.")
        if demod_freq != 0.0:
            acquisitions = self._demodulated_scope(frequency=-demod_freq, phase_offset=demod_phase_offset)
        if integrate:
            integrate_from, integrate_to = integration_range
            acquisitions = self._integrated_scope(
                scope_acquisitions=acquisitions, integrate_from=integrate_from, integrate_to=integrate_to
            )
        return acquisitions.scope.path0.data, acquisitions.scope.path1.data

    def counts_object(self) -> Counts:
        """Returns a Counts object containing the counts of each state.

        Returns:
            Counts: Counts object containing the counts of each state.
        Raises:
            NotImplementedError: this method is not implemented for n measurements on the same qubit
        """
        if sum(result["measurement"] for result in self.qblox_raw_results) != 0:
            raise NotImplementedError("Counts for multiple measurements on a single qubit are not supported")
        return self.qblox_bins_acquisitions.counts()

    def counts(self) -> dict:
        """Returns a Counts object containing the counts of each state.

        Returns:
            Counts: Counts object containing the counts of each state.
        Raises:
            NotImplementedError: this method is not implemented for n measurements on the same qubit
        """
        if sum(result["measurement"] for result in self.qblox_raw_results) != 0:
            raise NotImplementedError("Counts for multiple measurements on a single qubit are not supported")
        return self.qblox_bins_acquisitions.counts().as_dict()

    def samples(self) -> np.ndarray:
        """Returns an array containing the measured samples.
        The shape of the returned array is ``(# sequencers, # bins)``.

        Returns:
            np.ndarray: An array containing the measured samples (0 or 1).
        Raises:
            NotImplementedError: this method is not implemented for n measurements on the same qubit
        """
        if sum(result["measurement"] for result in self.qblox_raw_results) != 0:
            raise NotImplementedError("Samples for multiple measurements on a single qubit are not supported")
        return self.qblox_bins_acquisitions.samples()

    @property
    def array(self) -> np.ndarray:
        # Save array data
        if self.qblox_scope_acquisitions is not None:
            if sum(result["measurement"] for result in self.qblox_raw_results) != 0:
                raise NotImplementedError(
                    "Scope acquisition for multiple measurements on a single qubit are not supported"
                )
            # The dimensions of the array are: (2, N) where N is the length of the scope.
            path0 = self.qblox_scope_acquisitions.scope.path0.data
            path1 = self.qblox_scope_acquisitions.scope.path1.data
            return np.array([path0, path1])

        bins_len = [len(bins) for bins in self.qblox_bins_acquisitions.bins]
        # Check that all measurements have the same number of bins.
        if len(set(bins_len)) != 1:
            raise IndexError(
                f"All measurements must have the same number of bins to return an array. Obtained {len(bins_len)} "
                f"measurements with {bins_len} bins respectively."
            )
        # The dimensions of the array are the following: (#sequencers*#measurements, 2, #bins)
        # #measurements are the number of measurements done for a particular single qubit
        bins = [[result.integration.path0, result.integration.path1] for result in self.qblox_bins_acquisitions.bins]

        return np.array(bins[0] if len(bins) == 1 else bins)

    @property
    def shape(self) -> list[int]:
        """QbloxResult 'shape' property.

        Returns:
            list[int]: Shape of the acquisitions.
        """
        return list(self.acquisitions().shape)

    def to_dict(self) -> dict:
        """
        Returns:
            dict: Dictionary containing all the class information.
        """
        return {
            RUNCARD.NAME: self.name.value,
            QBLOXRESULT.INTEGRATION_LENGTHS: self.integration_lengths,
            QBLOXRESULT.QBLOX_RAW_RESULTS: self.qblox_raw_results,
        }



""" Tests for CoordinateDecompose"""

from qililab.utils import coordinate_decompose


class TestCoordinateDecompose:
    """Unit tests checking the Experiment attributes and methods"""

    def test_coordinate_decompose_trivial_case(self):
        """Test that coordinate decomposition works well for the trivial case: 1D to 1D"""
        original_array = list(range(10))
        original_size = len(original_array)
        new_dimension_shape = [original_size]
        for index in original_array:
            new_indices = coordinate_decompose(
                new_dimension_shape=new_dimension_shape, original_size=original_size, original_idx=index
            )
            assert new_indices[0] == index

    def test_coordinate_decompose_2d(self):
        """Test that coordinate decomposition works well for the trivial case: 1D to 2D"""
        original_array = list(range(2 * 5))
        original_size = len(original_array)
        new_dimension_shape = [2, 5]

        test_index = 7
        target_result = [1, 2]

        new_indices = coordinate_decompose(
            new_dimension_shape=new_dimension_shape, original_size=original_size, original_idx=test_index
        )

        assert all(target_result == new_indices)

    def test_coordinate_decompose_3d(self):
        """Test that coordinate decomposition works well for the trivial case: 1D to 3D"""
        original_array = list(range(2 * 3 * 5))
        original_size = len(original_array)
        new_dimension_shape = [2, 3, 5]

        test_index = 27
        target_result = [1, 2, 2]

        new_indices = coordinate_decompose(
            new_dimension_shape=new_dimension_shape, original_size=original_size, original_idx=test_index
        )

        assert all(target_result == new_indices)

    def test_coordinate_decompose_4d(self):
        """Test that coordinate decomposition works well for the trivial case: 1D to 3D"""
        original_array = list(range(2 * 2 * 2 * 2))
        original_size = len(original_array)
        new_dimension_shape = [2, 2, 2, 2]

        test_index = 5
        target_result = [0, 1, 0, 1]

        new_indices = coordinate_decompose(
            new_dimension_shape=new_dimension_shape, original_size=original_size, original_idx=test_index
        )

        assert all(target_result == new_indices)



""" Test Results """
from collections import Counter

import numpy as np
import pytest

from qililab.result.counts import Counts
from qililab.result.qprogram import QProgramResults
from qililab.result.qprogram.qblox_measurement_result import QbloxMeasurementResult
from qililab.result.qprogram.quantum_machines_measurement_result import QuantumMachinesMeasurementResult


@pytest.fixture(name="psi_plus_counts")
def psi_plus_counts_fixture() -> Counts:
    """Fixture of counts measuring the Psi+ Bell state"""
    counts = Counts(n_qubits=2)
    counter = Counter({"00": 500, "01": 12, "10": 12, "11": 500})
    counts._counter = counter
    counts._total_measurements = sum(counter.values())
    return counts


@pytest.fixture(name="phi_plus_counts")
def phi_plus_counts_fixture() -> Counts:
    """Fixture of counts measuring the Phi+ Bell state"""
    counts = Counts(n_qubits=2)
    counter = Counter({"00": 12, "01": 500, "10": 500, "11": 12})
    counts._counter = counter
    counts._total_measurements = sum(counter.values())
    return counts


@pytest.fixture(name="qblox_qprogram_results_probs")
def fixture_raw_measurement_data() -> QProgramResults:
    """Fixture to obtain proabilities from a QProgramResult"""
    results = QProgramResults()
    thresholds_q0 = [1.0, 1.0, 0.0, 1.0, 0.0]
    thresholds_q1 = [0.0, 1.0, 0.0, 0.0, 1.0]
    for th_q0, th_q1 in zip(thresholds_q0, thresholds_q1):
        results.append_result(
            bus="readout_q0",
            result=QbloxMeasurementResult(
                bus="readout_q0",
                raw_measurement_data={"bins": {"integration": {"path0": [0], "path1": [0]}, "threshold": [th_q0]}},
            ),
        )
        results.append_result(
            bus="readout_q1",
            result=QbloxMeasurementResult(
                bus="readout_q1",
                raw_measurement_data={"bins": {"integration": {"path0": [0], "path1": [0]}, "threshold": [th_q1]}},
            ),
        )
    return results


@pytest.fixture(name="qm_qprogram_results_probs")
def fixture_raw_measurement_data_qm() -> QProgramResults:
    """Fixture to obtain proabilities from a QProgramResult"""
    results = QProgramResults()

    res_0 = QuantumMachinesMeasurementResult(
        bus="readout", I=np.array([0.2, 1.3, 0.07, 1.0, 0.7]), Q=np.array([0.5, 1.4, 0.6, 1.02, 0.1])
    )
    res_1 = QuantumMachinesMeasurementResult(
        bus="readout", I=np.array([0.76, 1.2, 1.8, 0.4, 0.9]), Q=np.array([0.9, 1.0, 0.46, 2.6, 0.97])
    )
    res_0.set_classification_threshold(0.4)
    res_1.set_classification_threshold(0.5)

    results.append_result(bus="readout_q0", result=res_0)
    results.append_result(bus="readout_q1", result=res_1)

    return results


class TestsCounts:
    """Test `Counts` functionalities."""

    def test_str(self, psi_plus_counts: Counts):
        """Test `__str__` method."""
        assert str(psi_plus_counts) == "{'00': 500, '01': 12, '10': 12, '11': 500}"

    def test_verify_state_errors(self, psi_plus_counts: Counts):
        """Test _verify_state checks when adding a new state."""
        bad_n_qubits_state = "001"
        bad_basis_state = "35"

        bad_n_qubits_error_msg = (
            f"State '{bad_n_qubits_state}' can't be added to Counts object with {psi_plus_counts.n_qubits} qubits."
        )
        bad_basis_error_msg = f"A state can only be made of 0s and 1s, got state '{bad_basis_state}'."

        with pytest.raises(IndexError, match=bad_n_qubits_error_msg):
            psi_plus_counts.add_measurement(state=bad_n_qubits_state)

        with pytest.raises(ValueError, match=bad_basis_error_msg):
            psi_plus_counts.add_measurement(state=bad_basis_state)

    def test_bad_merge_error(self, psi_plus_counts: Counts):
        """Test that a Counts object can't be merged with another Counts object with different number of qubits."""
        one_qubit_counts = Counts(n_qubits=1)
        erorr_msg = f"Counts object with {one_qubit_counts.n_qubits} can't be added to Counts object with {psi_plus_counts.n_qubits} qubits."
        with pytest.raises(IndexError, match=erorr_msg):
            psi_plus_counts.add_counts(one_qubit_counts)

    def test_iadd(self, psi_plus_counts: Counts, phi_plus_counts: Counts):
        """Test `__iadd__` method."""
        psi_plus_counts += phi_plus_counts
        for count in psi_plus_counts.as_dict().values():
            assert count == 512

    def test_add_measurement(self, psi_plus_counts: Counts):
        """Test `add_measurement` method."""
        total_measurements_before = psi_plus_counts.total_measurements
        measurements_10_before = psi_plus_counts.as_dict()["10"]
        psi_plus_counts.add_measurement("10")
        assert psi_plus_counts.total_measurements == total_measurements_before + 1
        assert psi_plus_counts.as_dict()["10"] == measurements_10_before + 1


class TestProbabilitiesWithQProgram:
    """Class to test the `probabilities` method."""

    @pytest.mark.parametrize(
        "bus_mapping, expected_probs",
        [
            (None, {"00": 0.2, "01": 0.2, "10": 0.4, "11": 0.2}),
            (["readout_q1", "readout_q0"], {"00": 0.2, "01": 0.4, "10": 0.2, "11": 0.2}),
        ],
    )
    def test_probabilities_with_qblox(self, qblox_qprogram_results_probs: QProgramResults, bus_mapping, expected_probs):
        """Test probabilities are computed correctly with and without qubit mapping"""
        assert Counts.from_qprogram(qblox_qprogram_results_probs, bus_mapping).probabilities() == expected_probs

    @pytest.mark.parametrize(
        "bus_mapping, expected_probs",
        [
            (None, {"00": 0.0, "01": 0.4, "10": 0.2, "11": 0.4}),
            (["readout_q1", "readout_q0"], {"00": 0.0, "01": 0.2, "10": 0.4, "11": 0.4}),
        ],
    )
    def test_probabilities_with_quantum_machines(
        self, qm_qprogram_results_probs: QProgramResults, bus_mapping, expected_probs
    ):
        """Test probabilities are computed correctly with and without qubit mapping"""
        assert Counts.from_qprogram(qm_qprogram_results_probs, bus_mapping).probabilities() == expected_probs

    def test_probabilities_raises_empty(self):
        """Test an error is raised when the `qprogram_results` is empty"""
        qp_results = QProgramResults()
        with pytest.raises(ValueError) as empty_error:
            Counts.from_qprogram(qp_results)
        (msg,) = empty_error.value.args
        assert msg == f"Can not obtain counts with no measurments, {qp_results.__class__} empty"

    @pytest.mark.parametrize(
        "bus_mapping", [[], ["readout_q1", "readout_q1"], ["readout_q0", "readout_q1", "readout_q2"]]
    )
    def test_probabilities_raises_len_missmatch(self, qblox_qprogram_results_probs: QProgramResults, bus_mapping):
        """Test an error is raised when a mapping is not specified for all qubits"""
        n_qubits = len(qblox_qprogram_results_probs.results)
        unique_mapping = set(bus_mapping)
        with pytest.raises(ValueError) as len_missmatch:
            Counts.from_qprogram(qblox_qprogram_results_probs, bus_mapping)
        (msg,) = len_missmatch.value.args
        assert (
            msg
            == f"Expected mapping for all qubits. Results have {n_qubits} qubits, but only {len(unique_mapping)} diferent buses were specified on the mapping."
        )

    @pytest.mark.parametrize("bus_mapping", [["readout_q1", "readout_q2"], ["readout_q2", "readout_q3"]])
    def test_probabilities_raises_bus_missmatch(self, qblox_qprogram_results_probs: QProgramResults, bus_mapping):
        """Test an error is raised when not all buses are mapped into a qubit"""
        with pytest.raises(ValueError) as bus_missmatch:
            Counts.from_qprogram(qblox_qprogram_results_probs, bus_mapping)
        (msg,) = bus_missmatch.value.args
        assert (
            msg
            == "No measurements found for all specified buses, check the name of the buses provided with the mapping match all the buses specified in runcard."
        )



import numpy as np
import pytest
from scipy.special import erf

from qililab.waveforms import FlatTop


@pytest.fixture(name="flat_top")
def fixture_flat_top():
    return FlatTop(amplitude=1, duration=10, gaussian=0.2, buffer=1)


class TestGaussian:
    def test_init(self, flat_top):
        """Test __init__ method"""
        assert flat_top.amplitude == 1
        assert flat_top.duration == 10
        assert flat_top.gaussian == 0.2
        assert flat_top.buffer == 1

    def test_envelope(self, flat_top):
        """Test envelope method"""

        # calculate envelope
        x = np.arange(0, flat_top.duration, 2)
        A = flat_top.amplitude
        g = flat_top.gaussian
        buf = flat_top.buffer
        dur = flat_top.duration
        envelope = 0.5 * A * np.real((erf(g * x - buf) - erf(g * (x - (dur + -buf / g)))))

        assert np.allclose(flat_top.envelope(resolution=2), envelope)

    def test_get_duration(self, flat_top):
        """Test that the duration is retrieved properly."""
        assert flat_top.get_duration() == flat_top.duration



"""Unit tests for the Runcard class."""

import ast
import copy
import re
from dataclasses import asdict
from warnings import catch_warnings, simplefilter

import pytest

from qililab.constants import GATE_ALIAS_REGEX
from qililab.settings import Runcard, DigitalCompilationSettings, AnalogCompilationSettings
from qililab.settings.analog.flux_control_topology import FluxControlTopology
from qililab.settings.digital.gate_event_settings import GateEventSettings
from qililab.typings import Parameter
from tests.data import Galadriel


@pytest.fixture(name="runcard")
def fixture_runcard():
    return Runcard(**copy.deepcopy(Galadriel.runcard))


@pytest.fixture(name="digital")
def fixture_digital_compilation_settings(runcard: Runcard):
    return runcard.digital

@pytest.fixture(name="analog")
def fixture_analog_compilation_settings(runcard: Runcard):
    return runcard.analog


class TestRuncard:
    """Unit tests for the Runcard dataclass initialization."""

    def test_attributes(self, runcard: Runcard):
        """Test that the attributes of the Runcard are casted into dataclasses, and that
        the values they contain are the same as the input dictionaries."""

        assert isinstance(runcard.name, str)
        assert isinstance(runcard.instruments, list)
        assert isinstance(runcard.instruments[0], dict)
        assert isinstance(runcard.buses, list)
        assert isinstance(runcard.digital, DigitalCompilationSettings)
        assert isinstance(runcard.analog, AnalogCompilationSettings)


class TestDigitalCompilationSettings:
    """Unit tests for the DigitalCompilationSettings class."""

    def test_attributes(self, digital: DigitalCompilationSettings):
        """Test that the Runcard.GatesSettings dataclass contains the right attributes."""
        assert isinstance(digital.delay_before_readout, int)
        assert isinstance(digital.gates, dict)
        assert all(
            (isinstance(key, str), isinstance(event, GateEventSettings))
            for key, settings in digital.gates.items()
            for event in settings
        )

    def test_get_parameter_fails(self, digital: DigitalCompilationSettings):
        with pytest.raises(ValueError, match="Could not find gate alias in gate settings."):
            digital.get_parameter(alias="alias", parameter=Parameter.DURATION)

        with pytest.raises(ValueError):
            digital.get_parameter(alias="non-existent-bus", parameter=Parameter.DELAY)

    def test_get_gate(self, digital):
        """Test the ``get_gate`` method of the Runcard.GatesSettings class."""
        gates_qubits = [
            (re.search(GATE_ALIAS_REGEX, alias)["gate"], re.search(GATE_ALIAS_REGEX, alias)["qubits"])
            for alias in digital.gates.keys()
        ]
        assert all(
            isinstance(gate_event, GateEventSettings)
            for gate_name, gate_qubits in gates_qubits
            for gate_event in digital.get_gate(name=gate_name, qubits=ast.literal_eval(gate_qubits))
        )

        # check that CZs commute
        # CZ(0,1) doesn't have spaces in the tuple string
        assert isinstance(digital.get_gate(name="CZ", qubits=(1, 0))[0], GateEventSettings)
        assert isinstance(digital.get_gate(name="CZ", qubits=(0, 1))[0], GateEventSettings)

        # CZ(0, 2) has spaces in the tuple string
        assert isinstance(digital.get_gate(name="CZ", qubits=(2, 0))[0], GateEventSettings)
        assert isinstance(digital.get_gate(name="CZ", qubits=(0, 2))[0], GateEventSettings)

    def test_get_gate_raises_error(self, digital):
        """Test that the ``get_gate`` method raises an error when the name is not found."""
        name = "test"
        qubits = 0

        error_string = re.escape(f"Gate {name} for qubits {qubits} not found in settings").replace(
            "\\", ""
        )  # fixes re.escape bug
        with pytest.raises(KeyError, match=error_string):
            digital.get_gate(name, qubits=qubits)

    def test_gate_names(self, digital: DigitalCompilationSettings):
        """Test the ``gate_names`` method of the Runcard.GatesSettings class."""
        expected_names = list(digital.gates.keys())
        assert digital.gate_names == expected_names

    def test_set_platform_parameters(self, digital: DigitalCompilationSettings):
        """Test that with ``set_parameter`` we can change all settings of the platform."""
        digital.set_parameter(alias=None, parameter=Parameter.DELAY_BEFORE_READOUT, value=1234)
        assert digital.delay_before_readout == 1234

        digital.set_parameter(alias="drive_line_q0_bus", parameter=Parameter.DELAY, value=123)
        assert digital.buses["drive_line_q0_bus"].delay == 123

    def test_set_parameter_fails(self, digital: DigitalCompilationSettings):
        with pytest.raises(ValueError):
            digital.set_parameter(alias="non-existent-bus", parameter=Parameter.DELAY, value=123)

    @pytest.mark.parametrize("alias", ["X(0)", "M(0)"])
    def test_set_gate_parameters(self, alias: str, digital: DigitalCompilationSettings):
        """Test that with ``set_parameter`` we can change all settings of the platform's gates."""
        regex_match = re.search(GATE_ALIAS_REGEX, alias)
        assert regex_match is not None

        name = regex_match["gate"]
        qubits_str = regex_match["qubits"]
        qubits = ast.literal_eval(qubits_str)

        digital.set_parameter(alias=alias, parameter=Parameter.DURATION, value=1234)
        assert digital.get_gate(name=name, qubits=qubits)[0].pulse.duration == 1234

        digital.set_parameter(alias=alias, parameter=Parameter.PHASE, value=1234)
        assert digital.get_gate(name=name, qubits=qubits)[0].pulse.phase == 1234

        digital.set_parameter(alias=alias, parameter=Parameter.AMPLITUDE, value=1234)
        assert digital.get_gate(name=name, qubits=qubits)[0].pulse.amplitude == 1234

    @pytest.mark.parametrize("alias", ["X(0,)", "X()", "X", ""])
    def test_set_gate_parameters_raises_error_when_alias_has_incorrect_format(self, alias: str, digital):
        """Test that with ``set_parameter`` will raise error when alias has incorrect format"""
        with pytest.raises(ValueError, match=re.escape(f"Alias {alias} has incorrect format")):
            digital.set_parameter(alias=alias, parameter=Parameter.DURATION, value=1234)

class TestAnalogCompilationSettings:
    """Unit tests for the DigitalCompilationSettings class."""

    def test_attributes(self, analog: AnalogCompilationSettings):
        assert isinstance(analog.flux_control_topology, list)
        assert all(isinstance(topology, FluxControlTopology) for topology in analog.flux_control_topology)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from qililab.qprogram.operations.operation import Operation
from qililab.yaml import yaml

if TYPE_CHECKING:
    from qililab.qprogram.calibration import Calibration
    from qililab.qprogram.qprogram import QProgram
    from qililab.qprogram.variable import Variable


@yaml.register_class
class ExecuteQProgram(Operation):
    def __init__(
        self,
        qprogram: QProgram | Callable[..., QProgram],  # type: ignore[misc]
        bus_mapping: dict[str, str] | None = None,
        calibration: Calibration | None = None,
        debug: bool = False,
    ) -> None:
        super().__init__()
        self.qprogram: QProgram | Callable[[Variable, ...], QProgram] = qprogram  # type: ignore[misc]
        self.bus_mapping: dict[str, str] | None = bus_mapping
        self.calibration: Calibration | None = calibration
        self.debug: bool = debug



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Instrument Modules Utility Loader Module."""

from dataclasses import dataclass

from qililab.instruments.instruments import Instruments
from qililab.instruments.utils.instrument_reference import InstrumentReference


@dataclass
class Loader:
    """Loads the Instruments supported types froom the Instrument References"""

    def _get_instrument_or_raise_error_when_not_found_or_not_supported_type(self, instruments: Instruments, alias: str):
        """get instrument or raise error when not found"""
        instrument = instruments.get_instrument(alias=alias)
        if instrument is None:
            raise ValueError(f"No instrument object found for alias {alias}.")

        return instrument

    def replace_modules_from_settings_with_instrument_objects(
        self,
        instruments: Instruments,
        instrument_references: list[InstrumentReference],
    ):
        """Replace dictionaries from settings into its respective instrument classes.

        Args:
            instruments (Instruments): Instruments loaded into the platform.
            instrument_references (list[InstrumentReference]): List of references to the instruments
            with its alias to be retrieved from the Instrument Factory.

        Returns:
            list[Instrument]: List of the Instruments that manages the Controller with its device driver assigned.
        """
        return [
            self._get_instrument_or_raise_error_when_not_found_or_not_supported_type(
                instruments=instruments, alias=alias[1]
            )
            for alias, _ in instrument_references
        ]



import re
from unittest.mock import call, patch
import pytest
import networkx as nx
from qibo import Circuit, gates
from qibo.transpiler.optimizer import Preprocessing

from qibo.transpiler.placer import ReverseTraversal, StarConnectivityPlacer
from qibo.transpiler.router import Sabre, StarConnectivityRouter

from qililab.digital.circuit_router import CircuitRouter

# Different connectivities for testing the routing
linear_connectivity = [(0,1), (1,2), (2,3), (3,4)]
star_connectivity = [(0,1), (0,2), (0,3), (0,4)]

# Different topologies for testing the routing
linear_topology = nx.Graph(linear_connectivity)
star_topology = nx.Graph(star_connectivity)

# Linear circuit for testing specific routing
linear_circuit = Circuit(5)
linear_circuit.add(gates.H(0))
linear_circuit.add(gates.CNOT(0, 1))
linear_circuit.add(gates.CNOT(1, 2))
linear_circuit.add(gates.CNOT(2, 3))
linear_circuit.add(gates.CNOT(3, 4))

# Tests circuit
circuit_test = Circuit(5)
circuit_test.add(gates.H(0))
# Tests circuit with SWAP
circuit_w_swap_test = Circuit(5)
circuit_test.add(gates.H(0))
circuit_w_swap_test.add(gates.SWAP(0,1))
# Tests layouts
qibo_test_layout, test_layout = {1:0, 0:1}, [1, 0]
qibo_test_bad_layout = {0:0, 1:0}

#########################
### INTEGRATION TESTS ###
#########################
class TestCircuitRouterIntegration:
    """Tests for the circuit router class, integration tests."""

    @pytest.mark.parametrize(
        "type, topology",
        [("linear", linear_topology), ("star", star_topology), ("bad", linear_topology)]
    )
    def test_initialization(self, type, topology):
        """Test the initialization of the CircuitRouter class"""
        # Test the incorrect initialization of the CircuitRouter class
        if type == "bad":
            with pytest.raises(ValueError, match="StarConnectivity Placer and Router can only be used with star topologies"):
                _ = CircuitRouter(topology, router=StarConnectivityRouter)

        # Test the correct initialization of the CircuitRouter class
        elif type == "linear":
            router = CircuitRouter(topology, router=Sabre, placer=ReverseTraversal)
            assert isinstance(router.placer, ReverseTraversal)
            assert isinstance(router.router, Sabre)
        elif type == "star":
            router = CircuitRouter(topology, router=StarConnectivityRouter, placer=StarConnectivityPlacer)
            assert isinstance(router.placer, StarConnectivityPlacer)
            assert isinstance(router.router, StarConnectivityRouter)
            assert router.placer.middle_qubit is None
        if type in ["linear", "star"]:
            assert router.connectivity == topology
            assert isinstance(router.preprocessing, Preprocessing)

    def test_route_doesnt_affect_already_routed_circuit(self):
        """Test the routing of a circuit"""
        routed_circuit, final_layout = CircuitRouter(linear_topology).route(linear_circuit)

        assert final_layout == [0, 1, 2, 3, 4]
        assert routed_circuit.nqubits == linear_circuit.nqubits
        assert routed_circuit.depth == linear_circuit.depth
        assert [(gate.name, gate.qubits) for gate in routed_circuit.queue] == [(gate.name, gate.qubits)  for gate in linear_circuit.queue]

    def test_route_affects_non_routed_circuit(self):
        """Test the routing of a circuit"""

        routed_circuit, final_layout = CircuitRouter(star_topology).route(linear_circuit)

        # Assert that the circuit was routed:
        assert final_layout != [0, 1, 2, 3, 4]
        assert routed_circuit.nqubits == linear_circuit.nqubits
        assert routed_circuit.depth > linear_circuit.depth
        assert [(gate.name, gate.qubits) for gate in routed_circuit.queue] != [(gate.name, gate.qubits)  for gate in linear_circuit.queue]
        assert {gate.name for gate in routed_circuit.queue} >= {gate.name for gate in linear_circuit.queue} # Assert more gates


##################
### UNIT TESTS ###
##################
class TestCircuitRouterUnit:
    """Tests for the circuit router class, unit tests."""

    circuit_router = CircuitRouter(linear_topology)

    @pytest.mark.parametrize(
        "type, circuit, layout, least_swaps",
        [
            ("good", circuit_test, test_layout, 0),
            ("none_swaps", circuit_test, test_layout, None),
        ]
    )
    @patch("qililab.config.logger.info")
    @patch("qililab.digital.circuit_router.CircuitRouter._iterate_routing")
    def test_route(self, mock_iterate, mock_logger_info, type, circuit, layout, least_swaps):
        """Test the routing of a circuit."""
        # Set the mock returns to the parametrized test values:
        mock_iterate.return_value = (circuit, least_swaps, layout)

        # Execute the routing:
        if type in ["good", "none_swaps"]:
            routed_circuit, final_layout = self.circuit_router.route(linear_circuit)
            # Assert you return the same outputs as the mocked _iterate_routing
            assert (routed_circuit, final_layout) ==(circuit_test, test_layout)

        # Assert that the logger is called properly
        if type == "good":
            mock_logger_info.assert_has_calls([
                call("The best found routing, has 0 swaps."),
                call(f"{routed_circuit.wire_names}: Initial Re-mapping of the Original Logical Qubits (l_q), in the Physical Circuit: [l_q in wire 0, l_q in wire 1, ...]."),
                call(f"{final_layout}: Final Re-mapping (Initial + SWAPs routing) of the Original Logical Qubits (l_q), in the Physical Circuit: [l_q in wire 0, l_q in wire 1, ...].")
            ])
        elif type == "none_swaps":
            mock_logger_info.assert_called_once_with("No routing was done. Most probably due to routing iterations being 0.")

        # Assert that the routing pipeline was called with the correct arguments
        mock_iterate.assert_called_once_with(self.circuit_router.pipeline, linear_circuit, 10)

    @pytest.mark.parametrize("type, qibo_layout, layout", [("good", qibo_test_layout, test_layout), ("bad_layout", qibo_test_bad_layout, None)])
    def test_get_logical_qubit_of_each_wire(self, type, qibo_layout, layout):
        if type == "bad_layout":
            with pytest.raises(ValueError, match=re.escape(f"The final layout: {qibo_test_bad_layout} is not valid. i.e. a logical qubit is mapped to more than one physical qubit, or a key/value isn't a number. Try again, if the problem persists, try another placer/routing algorithm.")):
                _ = self.circuit_router._get_logical_qubit_of_each_wire(qibo_layout)

        elif type == "good":
            new_layout_format = self.circuit_router._get_logical_qubit_of_each_wire(qibo_layout)
            assert new_layout_format == layout

    @pytest.mark.parametrize(
        "type, circuit, qibo_layout, layout, least_swaps, iterations",
        [
            ("without_swaps", circuit_test, qibo_test_layout, test_layout, None, 10),
            ("with_swaps", circuit_w_swap_test, qibo_test_layout, test_layout, 1, 7),

        ]
    )
    @patch("qililab.digital.circuit_router.CircuitRouter._apply_initial_remap")
    @patch("qililab.digital.circuit_router.CircuitOptimizer.remove_redundant_start_controlled_gates")
    @patch("qililab.digital.circuit_router.Passes.__call__")
    def test_iterate_routing_with_and_without_swaps(self, mock_qibo_routing, mock_removing_swaps, mock_apply_initial_remap, type, circuit, qibo_layout, layout, least_swaps, iterations):
        """ Test the iterate routing of a circuit, with and without swaps."""
        # Add the mock return value to the parametrized test values:
        mock_qibo_routing.return_value = (circuit, qibo_layout)
        mock_removing_swaps.return_value = circuit
        mock_apply_initial_remap.return_value = circuit

        # Execute the iterate_routing:
        routed_circuit, least_swaps, final_layout = self.circuit_router._iterate_routing(self.circuit_router.pipeline, circuit, iterations)

        # Assert calls on the routing algorithm:
        if type == "with_swaps":
            # Assert called as many times as number of iterations, since there are swaps present:
            mock_qibo_routing.assert_has_calls([call(circuit)]*iterations)
            mock_removing_swaps.assert_has_calls([call(circuit, gates.SWAP)]*iterations)
            expected_least_swaps = 1
        elif type == "without_swaps":
            # Assert only called once, since there are no swaps:
            mock_qibo_routing.assert_called_once_with(circuit)
            mock_removing_swaps.assert_called_once_with(circuit, gates.SWAP)
            expected_least_swaps = iterations = 0 # Since there are no swaps, no iterations have been needed

        # Assert you return the correct outputs:
        assert (routed_circuit, least_swaps, final_layout) == (circuit, expected_least_swaps, layout)

    def test_if_star_algorithms_for_nonstar_connectivity(self):
        """Test the routing of a circuit"""
        circ_router = self.circuit_router

        # Assert cases where it needs to return True
        assert True == circ_router._if_star_algorithms_for_nonstar_connectivity(linear_topology, StarConnectivityPlacer(), circ_router.router)
        assert True == circ_router._if_star_algorithms_for_nonstar_connectivity(linear_topology, ReverseTraversal(circ_router.router), StarConnectivityRouter())
        assert True == circ_router._if_star_algorithms_for_nonstar_connectivity(linear_topology, StarConnectivityPlacer(), StarConnectivityRouter())

        # Assert cases where it needs to return False
        assert False == circ_router._if_star_algorithms_for_nonstar_connectivity(linear_topology, ReverseTraversal(circ_router.router), circ_router.router)
        assert False == circ_router._if_star_algorithms_for_nonstar_connectivity(star_topology, ReverseTraversal(circ_router.router), StarConnectivityRouter())
        assert False == circ_router._if_star_algorithms_for_nonstar_connectivity(star_topology, circ_router.placer, circ_router.router)

    def test_highest_degree_node(self):
        """Test the _highest_degree_node method."""
        # Test new edited linear topology
        edited_linear_topology = nx.Graph(linear_connectivity)
        edited_linear_topology.add_edges_from([(2,3),(2,4)])
        assert self.circuit_router._highest_degree_node(edited_linear_topology) == 2

        # Test the star topology with the 0 as central
        assert self.circuit_router._highest_degree_node(star_topology) == 0

    def test_if_layout_is_not_valid(self):
        """Test the _if_layout_is_not_valid method."""
        # Test valid layout
        assert not self.circuit_router._if_layout_is_not_valid(qibo_test_layout)

        # Test invalid layout
        assert self.circuit_router._if_layout_is_not_valid(qibo_test_bad_layout)

    @patch("qililab.digital.circuit_router.CircuitRouter._check_ReverseTraversal_routing_connectivity")
    @patch("qililab.digital.circuit_router.logger.warning")
    def test_build_placer(self, mock_logger_warning, mock_check_reverse):
        """Test the _build_placer method."""
        mock_check_reverse.return_value = (ReverseTraversal, {"routing_algorithm": self.circuit_router.router})

        # Test default placer (ReverseTraversal)
        placer = self.circuit_router._build_placer(None, self.circuit_router.router, linear_topology)
        assert isinstance(placer, ReverseTraversal)
        assert (placer.connectivity, placer.routing_algorithm) == (linear_topology, self.circuit_router.router)

        # Test ReverseTraversal placer
        placer = self.circuit_router._build_placer(ReverseTraversal, self.circuit_router.router, linear_topology)
        assert isinstance(placer, ReverseTraversal)
        assert placer.connectivity == linear_topology
        assert hasattr(placer, "routing_algorithm") == True

        # Test StarConnectivityPlacer
        placer = self.circuit_router._build_placer(StarConnectivityPlacer, self.circuit_router.router, star_topology)
        assert isinstance(placer, StarConnectivityPlacer)
        assert placer.middle_qubit is None
        assert hasattr(placer, "routing_algorithm") is False
        assert hasattr(placer, "connectivity") is True

        # Test ReverseTraversal with kwargs
        mock_check_reverse.reset_mock()
        placer = self.circuit_router._build_placer((ReverseTraversal, {"routing_algorithm": self.circuit_router.router}), self.circuit_router.router, linear_topology)
        mock_check_reverse.assert_called_once_with(ReverseTraversal, {"routing_algorithm": self.circuit_router.router}, linear_topology, self.circuit_router.router)
        assert isinstance(placer, ReverseTraversal)
        assert (placer.connectivity, placer.routing_algorithm) == (linear_topology, self.circuit_router.router)

        # Test invalid placer type
        with pytest.raises(TypeError, match="`placer` arg"):
            self.circuit_router._build_placer("invalid_placer", self.circuit_router.router, linear_topology)

        # Test Placer instance, instead than subclass:
        reverse_placer_instance = ReverseTraversal(self.circuit_router.router, linear_topology)
        mock_logger_warning.reset_mock()
        placer = self.circuit_router._build_placer(reverse_placer_instance, self.circuit_router.router, linear_topology)
        assert isinstance(placer, ReverseTraversal)
        assert placer.connectivity == linear_topology
        assert hasattr(placer, "routing_algorithm") == True
        # No call to warning, cause checking of Reverse is mocked.

        star_placer_instance = StarConnectivityPlacer(star_topology)
        mock_logger_warning.reset_mock()
        placer = self.circuit_router._build_placer(star_placer_instance, self.circuit_router.router, star_topology)
        assert isinstance(placer, StarConnectivityPlacer)
        assert placer.middle_qubit is None
        assert hasattr(placer, "routing_algorithm") is False
        assert hasattr(placer, "connectivity") == True
        mock_logger_warning.assert_called_once_with("Substituting the placer connectivity by the transpiler/platform/coupling_map one.")

        reverse_traversal_instance = ReverseTraversal(self.circuit_router.router, connectivity=linear_topology)
        placer = self.circuit_router._build_placer(reverse_traversal_instance, self.circuit_router.router, linear_topology)
        assert isinstance(placer, ReverseTraversal)
        assert (placer.connectivity, placer.routing_algorithm) == (linear_topology, self.circuit_router.router)

        # Test Placer instance, with kwargs:
        placer_instance = StarConnectivityPlacer()
        placer_kwargs = {"connectivity": 1}
        mock_logger_warning.reset_mock()
        router = self.circuit_router._build_placer((placer_instance,placer_kwargs),self.circuit_router.router, linear_topology)
        assert hasattr(router, "lookahead") == False
        mock_logger_warning.assert_has_calls([
            call("Ignoring placer kwargs, as the placer is already an instance."),
            call("Substituting the placer connectivity by the transpiler/platform/coupling_map one.")
        ])

    @patch("qililab.digital.circuit_router.logger.warning")
    def test_build_router(self, mock_logger_warning):
        """Test the _build_router method."""

        # Test default router (Sabre)
        router = self.circuit_router._build_router(None, linear_topology)
        assert isinstance(router, Sabre)
        assert router.connectivity == linear_topology

        # Test StarConnectivityRouter
        router = self.circuit_router._build_router(StarConnectivityRouter, star_topology)
        assert isinstance(router, StarConnectivityRouter)
        assert router.middle_qubit is None
        assert router.connectivity == star_topology

        # Test Sabre router with kwargs
        router = self.circuit_router._build_router((Sabre, {"lookahead": 2}), linear_topology)
        assert isinstance(router, Sabre)
        assert router.connectivity == linear_topology
        assert router.lookahead == 2

        # Test invalid router type
        with pytest.raises(TypeError, match="`router` arg"):
            self.circuit_router._build_router("invalid_router", linear_topology)

        # Test Router instance, instead of subclass
        sabre_instance = Sabre(linear_topology)
        mock_logger_warning.reset_mock()
        router = self.circuit_router._build_router(sabre_instance, linear_topology)
        assert isinstance(router, Sabre)
        assert router.connectivity == linear_topology
        mock_logger_warning.assert_has_calls([call("Substituting the router connectivity by the transpiler/platform/coupling_map one.")])


        star_router_instance = StarConnectivityRouter(star_topology)
        mock_logger_warning.reset_mock()
        router = self.circuit_router._build_router(star_router_instance, star_topology)
        assert isinstance(router, StarConnectivityRouter)
        assert router.middle_qubit is None
        assert router.connectivity == star_topology
        mock_logger_warning.assert_has_calls([call("Substituting the router connectivity by the transpiler/platform/coupling_map one.")])

        # Test Router instance, with kwargs:
        sabre_instance = Sabre(linear_topology, lookahead=2)
        router_kwargs = {"lookahead": 3}
        mock_logger_warning.reset_mock()
        router = self.circuit_router._build_router((sabre_instance,router_kwargs), linear_topology)
        assert router.lookahead == 2
        mock_logger_warning.assert_has_calls([call("Ignoring router kwargs, as the router is already an instance."),
            call("Substituting the router connectivity by the transpiler/platform/coupling_map one.")])

    def test_check_reverse_traversal_routing_connectivity(self):
        """Test the _check_ReverseTraversal_routing_connectivity method."""
        # Test for a linear topology
        og_placer = ReverseTraversal
        og_kwargs = {"routing_algorithm": Sabre}
        router = Sabre

        # Check with placer and routing algorithm both subclasses
        placer, kwargs = self.circuit_router._check_ReverseTraversal_routing_connectivity(og_placer, og_kwargs, linear_topology, router)
        assert (placer, kwargs) == (og_placer, og_kwargs)

        # Test for a weird router of the reversal
        og_kwargs = {"routing_algorithm": int, "lookahead": 2}
        with pytest.raises(TypeError, match=re.escape("`routing_algorithm` `Placer` kwarg (<class 'int'>) must be a `Router` subclass or instance, in `execute()`, `compile()`, `transpile_circuit()` or `route_circuit()`.")):
            self.circuit_router._check_ReverseTraversal_routing_connectivity(og_placer, og_kwargs, star_topology, router)

        # Test that the routing_algorithm get automatically inserted:
        og_kwargs = {"lookahead": 2}
        placer, kwargs = self.circuit_router._check_ReverseTraversal_routing_connectivity(og_placer, og_kwargs, linear_topology, router)
        assert (placer, kwargs) == (og_placer, og_kwargs | {"routing_algorithm": router})

        # Test instance of Router, change to chips topology:
        og_kwargs = {"routing_algorithm": Sabre(connectivity=None)}
        placer, kwargs = self.circuit_router._check_ReverseTraversal_routing_connectivity(og_placer, og_kwargs, linear_topology, router)
        og_kwargs["routing_algorithm"].connectivity = linear_topology
        assert (placer, kwargs) == (og_placer, og_kwargs)

        # Test subclass of Router, change to chips topology:
        og_kwargs = {"routing_algorithm": Sabre(connectivity=None)}
        og_placer = ReverseTraversal(connectivity=linear_topology, routing_algorithm=og_kwargs["routing_algorithm"])
        placer, kwargs = self.circuit_router._check_ReverseTraversal_routing_connectivity(og_placer, og_kwargs, linear_topology, router)
        assert og_placer.routing_algorithm.connectivity == linear_topology
        assert (placer, kwargs) == (og_placer, og_kwargs)



"""Tests for the BiasTeeCorrection distortion class."""
import itertools

import numpy as np
import pytest

from qililab.pulse import Pulse
from qililab.pulse.pulse_distortion import BiasTeeCorrection
from qililab.pulse.pulse_shape import SNZ, Cosine, Drag, Gaussian, Rectangular

# Parameters for the BiasTeeCorrection.
NORM_FACTOR = [0.7]
TAU_BIAS_TEE = [0.7, 1.3]

# Parameters of the Pulse and its envelope.
AMPLITUDE = [0, 0.8, -0.3, -2.1]
PHASE = [0, np.pi / 4]
DURATION = [48]
FREQUENCY = [0.7e9]
RESOLUTION = [1.0]
SHAPE = [Rectangular(), Cosine(), Gaussian(num_sigmas=4), Drag(num_sigmas=4, drag_coefficient=1.0), SNZ(b=0.2, t_phi=2)]


@pytest.fixture(
    name="pulse_distortion",
    params=[BiasTeeCorrection(tau_bias_tee=tau_bias_tee) for tau_bias_tee in TAU_BIAS_TEE],
)
def fixture_pulse_distortion(request: pytest.FixtureRequest) -> BiasTeeCorrection:
    """Fixture for the BiasTeeCorrection distortion class."""
    return request.param


@pytest.fixture(
    name="envelope",
    params=[
        Pulse(amplitude=amplitude, phase=phase, duration=duration, frequency=frequency, pulse_shape=shape).envelope(
            resolution=resolution
        )
        for amplitude, phase, duration, frequency, resolution, shape in itertools.product(
            AMPLITUDE, PHASE, DURATION, FREQUENCY, RESOLUTION, SHAPE
        )
    ],
)
def fixture_envelope(request: pytest.FixtureRequest) -> np.ndarray:
    """Fixture for an envelope."""
    return request.param


def return_corrected_envelopes_examples(
    pulse_distortion: BiasTeeCorrection, envelope: np.ndarray, norm_factors: list[float]
):
    """Helper function that returns examples of envelopes with & without auto_norm"""
    norm_corr_envelopes = [pulse_distortion.apply(envelope=envelope)]
    norm_corr_envelopes.append(
        BiasTeeCorrection(tau_bias_tee=1.3, norm_factor=norm_factors[0]).apply(envelope=norm_corr_envelopes[0])
    )
    norm_corr_envelopes.append(
        BiasTeeCorrection(tau_bias_tee=0.5, norm_factor=norm_factors[1]).apply(envelope=norm_corr_envelopes[1])
    )
    norm_corr_envelopes.append(BiasTeeCorrection(tau_bias_tee=0.9).apply(envelope=norm_corr_envelopes[2]))
    not_norm_corr_envelopes = [
        BiasTeeCorrection(tau_bias_tee=0.9, auto_norm=False).apply(envelope=norm_corr_envelopes[2])
    ]
    not_norm_corr_envelopes.append(
        BiasTeeCorrection(tau_bias_tee=0.7, auto_norm=False).apply(envelope=not_norm_corr_envelopes[0])
    )
    return norm_corr_envelopes, not_norm_corr_envelopes


class TestBiasTeeCorrection:
    """Unit tests checking the BiasTeeCorrection attributes and methods"""

    def test_apply(self, pulse_distortion: BiasTeeCorrection, envelope: np.ndarray):
        """Test for the envelope method."""
        norm_factors = [0.85, 0.15]
        norm_corr_envelopes, not_norm_corr_envelopes = return_corrected_envelopes_examples(
            pulse_distortion, envelope, norm_factors
        )

        # Basic checks
        for corr_envelope in norm_corr_envelopes + not_norm_corr_envelopes:
            assert corr_envelope is not None
            assert isinstance(corr_envelope, np.ndarray)
            assert len(envelope) == len(corr_envelope)
            assert (
                not np.array_equal(corr_envelope, envelope)
                or np.max(np.abs(np.real(envelope))) == np.max(np.abs(np.real(corr_envelope))) == 0.0
            )

        # Check that norm_factor and auto_norm is working properly
        assert (
            0.0  # Testing/Discarting the amplitude = 0 cases, for both maxs and mins.
            == round(np.max(np.abs(np.real(envelope))), 13)
            == round(np.min(np.abs(np.real(envelope))), 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[0]))), 13)
            == round(np.min(np.abs(np.real(norm_corr_envelopes[0]))), 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[1]))), 13)
            == round(np.min(np.abs(np.real(norm_corr_envelopes[1]))), 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[2]))), 13)
            == round(np.min(np.abs(np.real(norm_corr_envelopes[2]))), 13)
            == round(np.max(np.abs(np.real(not_norm_corr_envelopes[0]))), 13)
            == round(np.min(np.abs(np.real(not_norm_corr_envelopes[0]))), 13)
            == round(np.max(np.abs(np.real(not_norm_corr_envelopes[1]))), 13)
            == round(np.min(np.abs(np.real(not_norm_corr_envelopes[1]))), 13)
        ) or (
            # Actual testing that the norm_factors are working properly, the factors
            # should correspond to the ones in the function "return_corrected_envelopes_examples()"
            round(np.max(np.abs(np.real(norm_corr_envelopes[0]))), 14)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[1]))) / norm_factors[0], 14)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[2]))) / (norm_factors[0] * norm_factors[1]), 14)
            == round(np.max(np.abs(np.real(envelope))) * pulse_distortion.norm_factor, 14)
            # Testing that the auto_norm changes the norm from the previous
            != round(np.max(np.abs(np.real(not_norm_corr_envelopes[0]))) / (norm_factors[0] * norm_factors[1]), 14)
            # The next one has auto_norm = False again, so it has to be different again.
            != round(np.max(np.abs(np.real(not_norm_corr_envelopes[1]))) / (norm_factors[0] * norm_factors[1]), 14)
        )

    def test_from_dict(self, pulse_distortion: BiasTeeCorrection):
        """Test for the to_dict method."""
        dictionary = pulse_distortion.to_dict()
        pulse_distortions = [BiasTeeCorrection.from_dict(dictionary)]

        dictionary.pop("sampling_rate")
        pulse_distortions.append(BiasTeeCorrection.from_dict(dictionary))

        dictionary.pop("name")
        pulse_distortions.append(BiasTeeCorrection.from_dict(dictionary))

        dictionary["tau_bias_tee"] = 0.5
        dictionary["sampling_rate"] = 2.0
        pulse_distortions.append(BiasTeeCorrection.from_dict(dictionary))

        for distortion in pulse_distortions:
            assert distortion is not None
            assert isinstance(distortion, BiasTeeCorrection)

    def test_to_dict(self, pulse_distortion: BiasTeeCorrection):
        """Test for the to_dict method."""
        dictionary = pulse_distortion.to_dict()

        assert dictionary is not None
        assert isinstance(dictionary, dict)
        assert dictionary == {
            "name": pulse_distortion.name.value,
            "tau_bias_tee": pulse_distortion.tau_bias_tee,
            "sampling_rate": pulse_distortion.sampling_rate,
            "norm_factor": pulse_distortion.norm_factor,
            "auto_norm": pulse_distortion.auto_norm,
        }

    def test_serialization(self, pulse_distortion: BiasTeeCorrection):
        """Test that a serialization of the distortion is possible"""
        dictionary = pulse_distortion.to_dict()
        assert isinstance(dictionary, dict)

        new_pulse_distortion = BiasTeeCorrection.from_dict(dictionary)
        assert isinstance(new_pulse_distortion, BiasTeeCorrection)

        new_dictionary = new_pulse_distortion.to_dict()
        assert isinstance(new_dictionary, dict)
        assert new_dictionary == dictionary



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instrument Controller class"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from typing import Callable, Sequence, get_type_hints

from qililab.config import logger
from qililab.constants import INSTRUMENTCONTROLLER, RUNCARD
from qililab.instrument_connections.connection import Connection
from qililab.instruments.instrument import Instrument
from qililab.instruments.instruments import Instruments
from qililab.instruments.utils.instrument_reference import InstrumentReference
from qililab.instruments.utils.loader import Loader
from qililab.platform.components.bus_element import BusElement
from qililab.settings import Settings
from qililab.typings import ChannelID, Device, InstrumentControllerName, Parameter, ParameterValue
from qililab.utils import Factory


@dataclass(kw_only=True)
class InstrumentControllerSettings(Settings):
    """Contains the settings of a specific Instrument Controller.

    Args:
        connection (Connection): Connection class that represents the connection type of the Instrument Controller.
        modules: (list[InstrumentReference]): List of the Instrument References that links to the actual Instruments
                                                to be managed by the Instrument Controller.
        reset (bool, optional): Whether or not to reset the instrument after connecting to it. Defaults to True.
    """

    alias: str
    connection: Connection
    modules: list[InstrumentReference]
    reset: bool = True

    def __post_init__(self):
        super().__post_init__()
        if self.connection and isinstance(self.connection, dict):
            # Pop the connection name from the dictionary and instantiate its corresponding Connection class.
            self.connection = Factory.get(name=self.connection.pop(RUNCARD.NAME))(settings=self.connection)
        self.modules = [InstrumentReference.from_dict(settings=module) for module in self.modules]


class InstrumentController(BusElement, ABC):
    """Abstract base class declaring the necessary attributes
    and methods for the instrument controllers that manages
    the drivers (device) to one or several instruments and
    the connection to them.

    Args:
        name (InstrumentControllerName): name of the instrument controller
        settings (InstrumentControllerSettings): Settings of the instrument controller.
        device (Device): Driver instance of the instrument to operate the instrument controller.
        number_available_modules (int): Number of modules available in the Instrument Controller.
        modules (Sequence[Instrument]): Actual Instruments classes that manages the Instrument Controller.
        connected_modules_slot_ids (list[int]): List with the slot ids from the connected instruments
    """

    name: InstrumentControllerName
    settings: InstrumentControllerSettings  # a subtype of settings must be specified by the subclass
    device: Device  # a subtype of device must be specified by the subclass
    number_available_modules: int  # to be set by child classes
    modules: Sequence[Instrument]
    connected_modules_slot_ids: list[int]  # slot_id represents the number displayed in the cluster

    class CheckConnected:
        """Property used to check if the connection has established with an instrument."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, ref: "InstrumentController", *args, **kwargs):
            """
            Args:
                method (Callable): Class method.

            Raises:
                AttributeError: If connection has not been established with an instrument.
            """
            ref.connection.check_instrument_is_connected()
            return self._method(ref, *args, **kwargs)

    def __init__(self, settings: dict, loaded_instruments: Instruments):
        settings_class: type[InstrumentControllerSettings] = get_type_hints(self).get("settings")  # type: ignore
        self.settings = settings_class(**settings)
        self.modules = Loader().replace_modules_from_settings_with_instrument_objects(
            instruments=loaded_instruments,
            instrument_references=self.settings.modules,
        )
        if len(self.modules) <= 0:
            raise ValueError(f"The {self.name.value} Instrument Controller requires at least ONE module.")
        if len(self.modules) > self.number_available_modules:
            raise ValueError(
                f"The {self.name.value} Instrument Controller only supports {self.number_available_modules} module/s."
                + f"You have loaded {len(self.modules)} modules."
            )
        self.connected_modules_slot_ids = self._set_connected_modules_slot_ids()
        if len(self.connected_modules_slot_ids) > self.number_available_modules:
            raise ValueError(
                f"The {self.name.value} Instrument Controller only supports {self.number_available_modules} module/s."
                + f"You have connected {len(self.connected_modules_slot_ids)} modules."
            )
        if len(self.connected_modules_slot_ids) != len(self.modules):
            raise ValueError(
                f"The connected number of modules: {len(self.connected_modules_slot_ids)} differs from "
                + f"the available modules: {len(self.modules)}."
            )

    def _set_connected_modules_slot_ids(self) -> list[int]:
        """Initialize the modules slot ids from the settings"""
        return [instrument_reference.slot_id for instrument_reference in self.settings.modules]

    @abstractmethod
    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""

    @abstractmethod
    def _initialize_device(self):
        """Initialize device attribute to the corresponding device class."""

    @abstractmethod
    def _set_device_to_all_modules(self):
        """Sets the initialized device to all modules."""

    def _initialize_device_and_set_to_all_modules(self):
        """Initialize the Controller Device driver and sets it for all modules"""
        self._initialize_device()
        self._set_device_to_all_modules()

    def _release_device_to_all_modules(self):
        """Releases the device to all modules"""
        for module in self.modules:
            module.device = None

    def _release_device_and_set_to_all_modules(self):
        """Release the Controller Device driver and also for all modules"""
        self.device = None
        self._release_device_to_all_modules()

    def set_parameter(
        self,
        parameter: Parameter,
        value: ParameterValue,
        channel_id: ChannelID | None = None,
    ):
        """Updates the reset settings for the controller."""
        if parameter == Parameter.RESET:
            self.settings.reset = bool(value)
            return
        raise ValueError("Reset is the only property that can be set for an Instrument Controller.")

    def get_parameter(
        self,
        parameter: Parameter,
        channel_id: ChannelID | None = None,
    ):
        """Updates the reset settings for the controller."""
        if parameter == Parameter.RESET:
            return self.settings.reset
        raise ValueError("Reset is the only property that can be set for an Instrument Controller.")

    @CheckConnected
    def turn_on(self):
        """Turn on an instrument."""
        for module in self.modules:
            logger.info("Turn on instrument %s.", module.alias or module.name.value)
            module.turn_on()

    @CheckConnected
    def turn_off(self):
        """Turn off an instrument."""
        for module in self.modules:
            logger.info("Turn off instrument %s.", module.alias or module.name.value)
            module.turn_off()

    @CheckConnected
    def reset(self):
        """Reset instrument."""
        for module in self.modules:
            logger.info("Reset instrument %s.", module.alias or module.name.value)
            module.reset()

    @CheckConnected
    def initial_setup(self):
        """Initial setup of the instrument."""
        for module in self.modules:
            logger.info("Initial setup to instrument %s.", module.alias or module.name.value)
            module.initial_setup()

    def connect(self):
        """Establishes the connection with the instrument and performs a reset (if necessary)."""
        self._initialize_device_and_set_to_all_modules()
        self.connection.connect(device=self.device, device_name=str(self))
        if self.settings.reset:
            self.reset()

    def disconnect(self):
        """Resets the devices (if needed), close the connection to the instrument and releases the device."""
        if self.settings.reset:
            self.reset()
        self.connection.close()
        self._release_device_and_set_to_all_modules()

    @property
    def alias(self):
        """Instrument Controller 'alias' property.

        Returns:
            str: settings.alias.
        """
        return self.settings.alias

    @property
    def connection(self):
        """Instrument Controller 'connection' property.

        Returns:
            Connection: settings.connection.
        """
        return self.settings.connection

    @property
    def address(self):
        """Instrument Controller 'address' property.

        Returns:
            str: connection.address.
        """
        return self.connection.address

    def __str__(self):
        """String representation of an instrument controller."""
        return f"{self.alias}"

    def to_dict(self):
        """Return a dict representation of the InstrumentReference class."""
        return {
            RUNCARD.NAME: self.name.value,
            RUNCARD.ALIAS: self.alias,
            INSTRUMENTCONTROLLER.CONNECTION: self.connection.to_dict(),
            INSTRUMENTCONTROLLER.MODULES: [module.to_dict() for module in self.settings.modules],
            INSTRUMENTCONTROLLER.RESET: self.settings.reset,
        }



import pytest

from qililab import Parameter
from qililab.settings.digital.gate_event_settings import GateEventSettings


@pytest.fixture(name="schedule")
def fixture_schedule() -> list[dict]:
    """Returns a list of dictionary schedules"""

    return [
        {
            "bus": "drive_line_q0_bus",
            "pulse": {
                "amplitude": 0.8,
                "phase": 0,
                "duration": 200,
                "shape": {"name": "drag", "drag_coefficient": 0.8, "num_sigmas": 2},
            },
        },
        {
            "bus": "flux_line_q0_bus",
            "wait_time": 30,
            "pulse": {
                "amplitude": 0.8,
                "phase": 0,
                "duration": 200,
                "shape": {"name": "drag", "drag_coefficient": 0.8, "num_sigmas": 2},
                "options": {"test_option": 1},
            },
        },
        {
            "bus": "drive_line_q0_bus",
            "pulse": {
                "amplitude": 0.8,
                "phase": 0,
                "duration": 100,
                "shape": {"name": "rectangular"},
            },
        },
    ]


class TestGateEventSettings:
    def test_init(self, schedule):
        """ "Test init method"""
        gate_event = GateEventSettings(**schedule[1])
        assert gate_event.bus == "flux_line_q0_bus"
        assert gate_event.wait_time == 30
        # test pulse
        pulse = gate_event.pulse
        assert isinstance(pulse, GateEventSettings.GatePulseSettings)
        assert pulse.amplitude == 0.8
        assert pulse.phase == 0
        assert pulse.duration == 200
        assert pulse.shape == {"name": "drag", "drag_coefficient": 0.8, "num_sigmas": 2}
        assert pulse.options["test_option"] == 1

    def test_set_parameter(self, schedule):
        """Test the set parameter method"""
        gate_event = GateEventSettings(**schedule[1])
        gate_event.set_parameter(parameter=Parameter.WAIT_TIME, value=10)
        assert gate_event.wait_time == 10
        gate_event.set_parameter(parameter=Parameter.AMPLITUDE, value=10)
        assert gate_event.pulse.amplitude == 10
        gate_event.set_parameter(parameter=Parameter.DRAG_COEFFICIENT, value=10)
        assert gate_event.pulse.shape["drag_coefficient"] == 10



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""__init__.py"""

from .qblox_adc_sequencer import QbloxADCSequencer
from .qblox_d5a import QbloxD5a
from .qblox_draw import QbloxDraw
from .qblox_module import QbloxModule
from .qblox_qcm import QbloxQCM
from .qblox_qcm_rf import QbloxQCMRF
from .qblox_qrm import QbloxQRM
from .qblox_qrm_rf import QbloxQRMRF
from .qblox_s4g import QbloxS4g
from .qblox_sequencer import QbloxSequencer

__all__ = [
    "QbloxADCSequencer",
    "QbloxD5a",
    "QbloxDraw",
    "QbloxModule",
    "QbloxQCM",
    "QbloxQCMRF",
    "QbloxQRM",
    "QbloxQRMRF",
    "QbloxS4g",
    "QbloxSequencer",
]



import pytest
from unittest.mock import MagicMock
from qililab.platform.components.buses import Buses
from qililab.platform.components.bus import Bus

@pytest.fixture
def mock_buses():
    bus1 = MagicMock(spec=Bus)
    bus2 = MagicMock(spec=Bus)
    bus1.alias = "bus1"
    bus2.alias = "bus2"
    bus1.has_adc.return_value = False
    bus2.has_adc.return_value = True
    return [bus1, bus2]

@pytest.fixture
def buses(mock_buses):
    return Buses(elements=mock_buses)

def test_buses_initialization(buses, mock_buses):
    assert len(buses) == len(mock_buses)
    assert buses.elements == mock_buses

def test_buses_add(buses):
    new_bus = MagicMock(spec=Bus)
    new_bus.alias = "bus3"
    buses.add(new_bus)
    assert len(buses) == 3
    assert buses.get("bus3") == new_bus

def test_buses_get_existing(buses, mock_buses):
    bus = buses.get("bus1")
    assert bus == mock_buses[0]

def test_buses_get_non_existing(buses):
    bus = buses.get("non_existent_bus")
    assert bus is None

def test_buses_len(buses):
    assert len(buses) == 2

def test_buses_iter(buses, mock_buses):
    buses_iter = iter(buses)
    assert list(buses_iter) == mock_buses

def test_buses_getitem(buses, mock_buses):
    assert buses[0] == mock_buses[0]
    assert buses[1] == mock_buses[1]

def test_buses_to_dict(buses, mock_buses):
    mock_buses[0].to_dict.return_value = {"alias": "bus1"}
    mock_buses[1].to_dict.return_value = {"alias": "bus2"}
    expected_dict = [{"alias": "bus1"}, {"alias": "bus2"}]
    assert buses.to_dict() == expected_dict

def test_buses_str(buses):
    mock_buses = buses.elements
    expected_str = f"{str(mock_buses[0])}\n{str(mock_buses[1])}"
    assert str(buses) == expected_str

def test_buses_readout_buses(buses, mock_buses):
    readout_buses = buses.readout_buses
    assert len(readout_buses) == 1
    assert readout_buses[0] == mock_buses[1]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""__init__.py"""

from .asdict_factory import dict_factory
from .coordinate_decomposition import coordinate_decompose
from .dictionaries import merge_dictionaries
from .factory import Factory
from .hashing import hash_qpy_sequence, hash_qua_program
from .nested_data_class import nested_dataclass
from .nested_dict_iterator import nested_dict_to_pandas_dataframe
from .signal_processing import demodulate
from .singleton import Singleton, SingletonABC
from .waveforms import Waveforms

__all__ = [
    "Factory",
    "Singleton",
    "SingletonABC",
    "Waveforms",
    "coordinate_decompose",
    "demodulate",
    "dict_factory",
    "hash_qpy_sequence",
    "hash_qua_program",
    "merge_dictionaries",
    "nested_dataclass",
    "nested_dict_to_pandas_dataframe",
]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SNZ pulse shape."""

from dataclasses import dataclass

import numpy as np

from qililab.config import logger
from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.typings import PulseShapeName
from qililab.utils import Factory


@Factory.register
@dataclass(frozen=True, eq=True)
class SNZ(PulseShape):
    """Sudden net zero pulse shape. It is composed of a half-duration positive rectangular pulse, followed
    by three stops to cross height = 0, to then have another half-duration negative rectangular pulse.

    More concretely, the shape from left to right is composed by:
        - half-duration positive rectangular pulse.
        - instantaneous stop at height b.
        - t-phi duration at height = 0.
        - instantaneous stop at height -b.
        - half-duration negative rectangular pulse.

    Examples:
        To get the envelope of a SNZ shape, with ``b`` and ``t_phi`` equal to ``B`` and ``T``, you need to do:

        .. code-block:: python

            from qililab.pulse.pulse_shape import SNZ

            snz_envelope = SNZ(b=B, t_phi=T).envelope(amplitude=1, duration=50)

        which for ``b=0.2``, ``t_phi=2`` and ``b=0.5``, ``t_phi=10``, look respectively like:

        .. image:: /classes_images/snzs.png
            :width: 800
            :align: center

    References:
        High-fidelity controlled-Z gate with maximal intermediate leakage operating at the speed
        limit in a superconducting quantum processor: https://arxiv.org/abs/2008.07411

    Args:
        b (float): Instant stops height when going from the rectangular half-duration to `height = 0`.
        t_phi (int): Time at `height = 0`, in the middle of the positive and negative rectangular pulses.
    """

    name = PulseShapeName.SNZ  # noqa: RUF045
    b: float  #: Instant stops height.
    t_phi: int  #: Time at `height = 0`.

    def __post_init__(self):
        # ensure t_phi is an int
        if not isinstance(self.t_phi, int):
            raise TypeError("t_phi for pulse SNZ has to be an integer. Since min time resolution is 1ns")

    def envelope(self, duration: int, amplitude: float, resolution: float = 1.0) -> np.ndarray:
        """Constant amplitude envelope.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            ndarray: Amplitude of the envelope for each time step.

        The duration of the each half-pulse is determined by the total pulse duration. Thus
        halfpulse_t = (duration - t_phi - 2) / 2. This implies that (duration - t_phi) should be even.
        The -2 in the formula above is due to the 2 impulses b.
        """
        # calculate the halfpulse duration
        halfpulse_t = (duration - 2 - self.t_phi) / 2
        halfpulse_t = int(halfpulse_t / resolution)

        envelope = np.zeros(round(duration / resolution))
        # raise warning if we are rounding
        if (duration / resolution) % 1 != 0 or (halfpulse_t / resolution) % 1 != 0:
            logger.warning(  # pragma: no cover
                f"Envelope length rounded to nearest value {len(envelope)} from division full_snz_duration ({duration}) / resolution ({resolution}) = {duration / resolution}"
            )
        envelope[:halfpulse_t] = amplitude * np.ones(halfpulse_t)  # positive square halfpulse
        envelope[halfpulse_t] = self.b * amplitude  # impulse b
        envelope[halfpulse_t + 2 + self.t_phi :] = 0  # t_phi
        envelope[halfpulse_t + 1 + self.t_phi] = -self.b * amplitude  # impulse -b
        envelope[halfpulse_t + 2 + self.t_phi :] = -amplitude * np.ones(halfpulse_t)  # negative square halfpulse

        return envelope



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ramp waveform."""

import numpy as np

from qililab.qprogram.decorators import requires_domain
from qililab.qprogram.variable import Domain
from qililab.yaml import yaml

from .waveform import Waveform


@yaml.register_class
class Ramp(Waveform):
    """Ramp waveform. The waveform's amplitude changes linearly from ``from_amplitude`` to ``to_amplitude`` in ``duration`` nanoseconds.

    Args:
        from_amplitude (float): The amplitude of the waveform at t=0.
        to_amplitude (float): The amplitude of the waveform at t=duration.
        duration (int): Duration of the waveform (ns).

    Examples:
        In the following example the waveform's amplitude will ramp from 0.0 to 1.0 in 50 ns.

        .. code-block:: python

            import qililab as ql

            # Create the waveform
            ramp_wf = ql.Ramp(from_amplitude=0.0, to_amplitude=1.0, duration=50)

            # Get waveform's envelope
            envelope = ramp_wf.envelope()
    """

    @requires_domain("from_amplitude", Domain.Voltage)
    @requires_domain("to_amplitude", Domain.Voltage)
    @requires_domain("duration", Domain.Time)
    def __init__(self, from_amplitude: float, to_amplitude: float, duration: int):
        super().__init__()
        self.from_amplitude = from_amplitude
        self.to_amplitude = to_amplitude
        self.duration = duration

    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Constant amplitude envelope.

        Args:
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            np.ndarray: Height of the envelope for each time step.
        """
        return np.linspace(self.from_amplitude, self.to_amplitude, round(self.duration / resolution))

    def get_duration(self) -> int:
        """Get the duration of the waveform.

        Returns:
            int: The duration of the waveform in ns.
        """
        return self.duration



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from qililab.yaml import yaml


@yaml.register_class
class CrosstalkMatrix:
    """A class to represent a crosstalk matrix where each index corresponds to a bus."""

    def __init__(self) -> None:
        """Initializes an empty crosstalk matrix."""
        self.matrix: dict[str, dict[str, float]] = {}
        self.flux_offsets: dict[str, float] = {}

    def to_array(self) -> np.ndarray:
        """Returns the np.array representation of the crosstalk matrix.

        Returns:
            np.ndarray: crosstalk matrix as a numpy array.
        """
        buses: set[str] = set(self.matrix.keys())
        for values in self.matrix.values():
            buses.update(values.keys())

        sorted_buses = sorted(buses)

        xtalk_matrix = np.empty(shape=[len(self.matrix), len(self.matrix)])
        for i, bus1 in enumerate(sorted_buses):
            for j, bus2 in enumerate(sorted_buses):
                xtalk_matrix[i, j] = self.matrix[bus1][bus2]
        return xtalk_matrix

    def inverse(self) -> "CrosstalkMatrix":
        """Returns the inverse version of the crosstalk matrix (as a bus dictionary).

        Returns:
            CrosstalkMatrix: inverse crosstalk matrix
        """
        inverse_xtalk_array = np.linalg.inv(self.to_array())
        return self.from_array(list(self.matrix.keys()), inverse_xtalk_array)

    def __getitem__(self, bus: str) -> dict[str, float]:
        """Returns the dictionary of crosstalk values for the given bus.

        Args:
            bus (str): The bus for which to get the crosstalk values.

        Returns:
            dict[str, float]: A dictionary of crosstalk values for the given bus.
        """
        if bus not in self.matrix:
            self.matrix[bus] = {}
        return self.matrix[bus]

    def __setitem__(self, key: str, value: dict[str, float]) -> None:
        """Sets the crosstalk values for the given bus.

        Args:
            key (str): The bus for which to set the crosstalk values.
            value (dict[str, float]): A dictionary of crosstalk values to set for the given bus.
        """
        if key not in self.matrix:
            self.matrix[key] = value
        else:
            for bus in value.keys():
                self.matrix[key][bus] = value[bus]

        if key not in self.flux_offsets:
            self.flux_offsets[key] = 0.0

    def __repr__(self) -> str:
        """Returns a string representation of the CrosstalkMatrix.

        Returns:
            str: A string representation of the CrosstalkMatrix.
        """
        return f"CrosstalkMatrix({self.matrix})"

    def __str__(self) -> str:
        """Produces a string representation of the matrix with rows and columns, labels included.
        The diagonal contains "\", since there is no value there, and if a value doesn't exist, the default 1.0 is shown.

        Returns:
            str: A string representation of the crosstalk matrix.
        """
        buses: set[str] = set(self.matrix.keys())
        for values in self.matrix.values():
            buses.update(values.keys())

        sorted_buses = sorted(buses)
        col_width = max(len(bus) for bus in sorted_buses) + 4  # Determine column width
        header = " " * col_width + " ".join(f"{bus:>{col_width}}" for bus in sorted_buses) + "\n"
        rows = []
        for bus1 in sorted_buses:
            row = [f"{bus1:{col_width}}"]
            for bus2 in sorted_buses:
                row.append(f"{self.matrix.get(bus1, {}).get(bus2, 1.0):{col_width}}")
            rows.append(" ".join(row))
        return header + "\n".join(rows)

    def set_offset(self, offset: dict[str, float]):
        """Modifies the offset based on the given bus and offset value

        Args:
            offset (dict[str, float]): dictionary containing the buses of the offsets to be added or modified and the value of said offsets
        """
        for bus in offset:
            self.flux_offsets[bus] = offset[bus]

    @classmethod
    def from_array(cls, buses: list[str], matrix_array: np.ndarray) -> "CrosstalkMatrix":
        """Creates crosstalk matrix from an array and corresponding set of buses. For a set of buses
        [bus1,bus2,...,busN] the corresponding matrix should have the same indices for rows and columns
        i.e. for the set of buses [bus1,bus2,...,busN], matrix[0,0] will be the coefficient for bus1[bus1],
        matrix[0,1] will be bus1[bus2], etc.

        Args:
            buses (list[str]): ordered list of buses for the crosstalk matrix
            matrix_array (np.ndarray): crosstalk matrix numpy array

        Returns:
            CrosstalkMatrix: CrosstalkMatrix: An instance of CrosstalkMatrix
        """

        instance = cls()
        instance.matrix = {}
        for i, bus1 in enumerate(buses):
            for j, bus2 in enumerate(buses):
                if bus1 not in instance.matrix:
                    instance.matrix[bus1] = {}
                instance.matrix[bus1][bus2] = float(matrix_array[i, j])

        if not instance.flux_offsets:
            for bus in buses:
                instance.flux_offsets[bus] = 0.0
        return instance

    @classmethod
    def from_buses(cls, buses: dict[str, dict[str, float]]) -> "CrosstalkMatrix":
        """
        Creates a CrosstalkMatrix with all possible associations set to 1.0.

        Args:
            buses (Sequence[str]): A sequence of bus names.

        Returns:
            CrosstalkMatrix: An instance of CrosstalkMatrix with all associations set to 1.0.
        """
        instance = cls()
        instance.matrix = {}
        instance.matrix = {bus1: {bus2: buses[bus1][bus2] for bus2 in buses} for bus1 in buses}

        if not instance.flux_offsets:
            for bus in buses:
                instance.flux_offsets[bus] = 0.0
        return instance


@yaml.register_class
class FluxVector:
    """Class to represent a flux vector. This is a dictionary of bus[flux] values"""

    def __init__(self) -> None:
        self.flux_vector: dict[str, float] = {}
        self.bias_vector: dict[str, float] = {}
        self.crosstalk: CrosstalkMatrix | None = None
        self.crosstalk_inverse: CrosstalkMatrix | None = None

    def __getitem__(self, bus: str) -> float:
        """Given a bus, returns its corresponding flux

        Args:
            bus (str): The bus for which to get the flux values.

        Returns:
            float: Flux value for the given bus
        """
        if self.bias_vector:
            return self.bias_vector[bus]
        return self.flux_vector[bus]

    def __setitem__(self, key: str, flux: float) -> None:
        """Given a bus, sets a new flux

        Args:
            key (str): The bus for which to set the flux values.
            flux (float): The new value for the given flux.

        """

        self.flux_vector[key] = flux
        if self.crosstalk:
            self.set_crosstalk(self.crosstalk)

    def set_crosstalk(self, crosstalk: CrosstalkMatrix):
        """Set the crosstalk compensation on the existing flux vector. This function does the matrix product to calculate the correct flux

        Args:
            crosstalk (CrosstalkMatrix): _description_

        Returns:
            _type_: _description_
        """
        self.crosstalk = crosstalk
        self.crosstalk_inverse = crosstalk.inverse()
        self.crosstalk_inverse.flux_offsets = self.crosstalk.flux_offsets

        self.bias_vector = self.flux_vector.copy()

        for bus_1 in self.crosstalk_inverse.matrix.keys():
            self.bias_vector[bus_1] = sum(
                (self.flux_vector[bus_2] + self.crosstalk_inverse.flux_offsets[bus_2])
                * self.crosstalk_inverse.matrix[bus_1][bus_2]
                for bus_2 in self.crosstalk_inverse.matrix[bus_1].keys()
            )

        return self.bias_vector

    def set_crosstalk_from_bias(self, crosstalk: CrosstalkMatrix):
        """Set the crosstalk compensation on the existing flux vector. This function does the matrix product to calculate the correct flux

        Args:
            crosstalk (CrosstalkMatrix): _description_

        """
        self.crosstalk = crosstalk
        self.crosstalk_inverse = crosstalk.inverse()

        self.bias_vector = self.flux_vector.copy()

        for bus_1 in self.crosstalk_inverse.matrix.keys():
            self.bias_vector[bus_1] = sum(
                self.flux_vector[bus_2] * self.crosstalk_inverse.matrix[bus_1][bus_2]
                for bus_2 in self.crosstalk_inverse.matrix[bus_1].keys()
            )

        return self.bias_vector

    def to_dict(self):
        """To dictionary method, returns the vector's dictionary

        Returns:
            dict[str, float]: Flux vector dictionary
        """
        if self.bias_vector:
            return self.bias_vector
        return self.flux_vector

    @classmethod
    def from_dict(cls, flux_dict: dict[str, float]) -> "FluxVector":
        """Creates a FluxVector instance from a dictionary of bus[flux]

        Args:
            flux_dict (dict[str,float]): Dictionary containing buses as keys and fluxes as values

        Returns:
            FluxVector: FluxVector instance
        """
        instance = cls()
        instance.flux_vector = flux_dict
        return instance



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass


@dataclass
class FluxControlTopology:
    """Dataclass fluxes (e.g. phix_q0 for phix control of qubit 0) and their corresponding bus (e.g. flux_line_q0_x)"""

    flux: str
    bus: str

    def to_dict(self):
        """Method to convert to dictionary"""
        return asdict(self)



import pytest
from qililab.settings.analog.flux_control_topology import FluxControlTopology

@pytest.fixture(name="topology")
def fixture_topology() -> FluxControlTopology:
    return FluxControlTopology(flux="flux_0", bus="flux_bus_q0")

class TestFluxControlTopology:
    def test_init(self, topology: FluxControlTopology):
        topology.flux == "flux_0"
        topology.bus == "flux_bus_q0"

    def test_to_dict_method(self, topology: FluxControlTopology):
        as_dict = topology.to_dict()
        assert isinstance(as_dict, dict)
        assert as_dict["flux"] == "flux_0"
        assert as_dict["bus"] == "flux_bus_q0"
